import os
import json

# fix for mac osx
os.environ['LC_ALL'] = 'en_US.UTF-8'
os.environ['LANG'] = 'en_US.UTF-8'

import click
from shutil import copyfile
from mlapp.mlapp_cli.common.cli_utilities import create_directory, create_file
from mlapp.mlapp_cli.common.files import amlignore_file, azureml_env_file, default_config_file
from mlapp.utils.general import get_project_root
from mlapp.integrations.aml.cli_help import cli_aml_help
from mlapp.integrations.aml.utils.workspace import init_workspace
from mlapp.integrations.aml.utils.datastore import get_datastore
from mlapp.integrations.aml.scripts.deploy_model import run_script as run_deploy_model_script
from mlapp.integrations.aml.scripts.publish_pipeline_endpoint import run_script as publish_pipeline_endpoint_script
from mlapp.integrations.aml.scripts.publish_multisteps_pipeline import run_script as \
    publish_multisteps_pipeline_script
from mlapp.integrations.aml.utils.cli_steps import steps
from mlapp.mlapp_cli.common.cli_utilities import clean_spaces
from azureml.core import Workspace

# constants
EMPTY_INPUT = ''

init_files_aml = {
    'aml_deployment.py': {
        'dir': 'deployment'
    },
    'aml_target_compute.py': {
        'dir': 'deployment'
    },
    'aml_flow.py': {
        'dir': 'deployment'
    }
}


@click.group("aml")
def commands():
    """
    MLApp AML Command

    Use it to run Azure Machine Learning commands.
    type --help on aml command to get more information.

    """
    pass


@commands.command("setup", help=cli_aml_help.get('setup'))
@click.option("-f", "--force", is_flag=True, default=False,
              help="Flag force setup if some of the AML files already exists in your project.")
def setup(force):
    if not force:
        is_initiated = False
        for file_name in init_files_aml:
            file_options = init_files_aml.get(file_name, file_name)
            directory = file_options.get('dir', 'root')
            if directory == 'root':
                full_path = os.path.join(os.getcwd(), file_name)
            else:
                if os.path.isdir(directory):
                    full_path = os.path.join(os.getcwd(), directory, file_name)
                else:
                    continue
            if os.path.exists(full_path):
                is_initiated = True
                break

        if is_initiated:
            click.secho(
                "ERROR: " + file_name + " already exists.\nHint: you can use 'mlapp aml setup --force' option to force setup (caution: force may override exsiting files).",
                fg='red')
            return

        # setup azure machine learning deployment files
        _setup_aml()
    else:
        _setup_aml()


def _setup_aml():
    # creates the deployment directory if not exists.
    create_directory(directory_name='deployment', include_init=False)

    # creates the environment directory if not exists.
    create_directory(directory_name='env', include_init=False)
    create_file(file_name='.env', path='env', content=azureml_env_file)

    # sets config.py environment file
    default_env_filename = ''
    config_file_content = default_config_file.replace("<FILENAME>", default_env_filename)
    create_file(file_name='config.py', content=config_file_content)

    for file_name in init_files_aml:
        file_options = init_files_aml.get(file_name, file_name)
        directory = file_options.get('dir', 'root')
        src = os.path.join(get_project_root(), 'mlapp', 'integrations', 'aml', 'generated_files', file_name)
        if directory == 'root':
            dst = os.path.join(os.getcwd(), file_name)
        else:
            dst = os.path.join(os.getcwd(), directory, file_name)

        # copy file from src to dst
        copyfile(src, dst)

    # creates amlignore
    create_file('.amlignore', content=amlignore_file)


@commands.command("deploy-model", help=cli_aml_help.get('deploy_model'))
@click.argument("experiment-name", required=True)
@click.argument("asset-name", required=True)
@click.option('-as', '--asset-label', default=None, help="Use it to add a label to your asset.")
@click.option('-rid', '--run-id', default=None, help="Use it to deploy a specific model.")
@click.option('-smetric', '--score-metric', default=None,
              help="Use it to choose best model according to a score metric (must be passed together with grater-is-better option).")
@click.option('-g', '--greater-is-better', default=False,
              help="Use it to set your score metric options (must be passed together with score-metric option).",
              type=bool)
@click.option('-cpu', '--cpu-cores', default=1, help="Use it to set number of cores in compute target machine.",
              type=int)
@click.option('-mgb', '--memory-gb', default=8, help="Use it to set memory size in compute target machine.", type=int)
def deploy_model(experiment_name, asset_name, asset_label, run_id, score_metric, greater_is_better, cpu_cores,
                 memory_gb):
    try:
        ws, _ = _get_aml_objects()
        run_deploy_model_script(ws, experiment_name, asset_name, asset_label, run_id, score_metric,
                                greater_is_better, cpu_cores, memory_gb)
    except Exception as e:
        click.secho(str(e), fg='red')


@commands.command("publish-pipeline", help=cli_aml_help.get('publish_pipeline'))
@click.argument("pipeline-name", required=True)
@click.argument("compute-target", required=True)
@click.option('-vs', '--vm-size', default='STANDARD_D2_V2', help="Use it to set vm size.")
@click.option('-mnn', '--min-nodes', default=0, help="Use it set min nodes number.", type=int)
@click.option('-mxn', '--max-nodes', default=4, help="Use it set max nodes number.", type=int)
def publish_pipeline(pipeline_name, compute_target, vm_size, min_nodes, max_nodes):
    try:
        ws, datastore = _get_aml_objects()
        publish_pipeline_endpoint_script(ws, datastore, pipeline_name, compute_target, vm_size, min_nodes,
                                         max_nodes)
    except Exception as e:
        click.secho(str(e), fg='red')


@commands.command("publish-multisteps-pipeline", help=cli_aml_help.get('publish_pipeline'))
@click.argument("pipeline-name", required=True)
def publish_multisteps_pipeline(pipeline_name):
    try:
        ws, datastore = _get_aml_objects()
    except:
        click.secho("ERROR: Please run 'mlapp init -aml' first or check your azureml Workspace credentials in config.py", fg='red')
        return

    number_of_steps_body = steps.get('number_of_steps', {})
    dependencies = number_of_steps_body.get('nested_dependency', {})
    number_of_steps = _display_message(number_of_steps_body)
    new_compute_targets = {}
    is_new_compue_target = False
    instructions = []
    for i in range(number_of_steps):
        args = {}
        is_default_values = False  # if true takes all default values for the test of the parameters.
        print('Please enter step %s arguments:' % str(i + 1))
        for dependency_name, body in dependencies.items():
            if is_default_values:
                args[dependency_name] = body.get('default', None)
            else:
                args[dependency_name] = _display_message(body)

            if dependency_name == 'compute_target':
                if args[dependency_name] in ws.compute_targets.keys():
                    break
                elif args[dependency_name] in new_compute_targets.keys():
                    args = new_compute_targets[args[dependency_name]]
                    break
                else:
                    click.echo('Found a new compute-target')
                    is_new_compue_target = True
            elif dependency_name == 'set_advanced' and not args[dependency_name]:
                # break if user don't want to set advanced settings
                is_default_values = True

        if is_new_compue_target:
            new_compute_targets[args['compute_target']] = args
            is_new_compue_target = False
        instructions.append(args)

    publish_multisteps_pipeline_script(ws, datastore, pipeline_name, instructions)


def _update_env_name(env_name):
    try:
        with open(os.path.join(os.getcwd(), 'config.py'), 'r') as f:
            config_content = f.read()
            exec(config_content)
            d = eval("settings")

        if 'aml' not in d:
            d['aml'] = {}

        d['aml']['environment'] = env_name

        config_file_content = '''settings = ''' + json.dumps(d, indent=2)
        create_file(file_name='config.py', content=config_file_content)

    except Exception as e:
        raise e


def _get_aml_objects():
    try:
        with open(os.path.join(os.getcwd(), 'config.py'), 'r') as f:
            config_content = f.read()
            exec(config_content)
            d: dict = eval("settings")

        # check if Workspace credentials exists
        aml = d.get('aml', {})
        if not aml:
            raise Exception("ERROR: please add AML Workspace credentials in your config.py under aml parameter.")

        tenant_id = aml.get('tenant_id')
        subscription_id = aml.get('subscription_id')
        resource_group = aml.get('resource_group')
        workspace_name = aml.get('workspace_name')
        datastore_name = aml.get('datastore_name', 'workspaceblobstore')

        if subscription_id is not None and resource_group is not None and workspace_name is not None:
            ws: Workspace = init_workspace(tenant_id, subscription_id, resource_group, workspace_name)
            datastore = get_datastore(ws, datastore_name)
            return ws, datastore
        else:
            raise Exception(
                "ERROR: credentials must include properties: subscription_id, resource_group, workspace_name.")
    except Exception as e:
        raise e


def _display_message(body):
    while True:
        display_name = body.get('display_name', '')
        short_description = body.get('short_description', '')

        # create message to display on the terminal.
        message = display_name + " (" + short_description + "): " if short_description != '' else display_name + ": "

        # get user input
        user_input = clean_spaces(input(message))

        # check for default and required values
        default_value = body.get('default', None)
        is_required = body.get('required', False)

        try:
            if default_value is not None and user_input == EMPTY_INPUT:
                user_input = default_value
            elif user_input == EMPTY_INPUT and is_required:
                raise Exception("'" + display_name + "' is required, please enter a valid value.")
            else:
                # validates user input
                validations_methods = body.get('validations', [])
                ans = True
                for validate in validations_methods:
                    ans &= validate(user_input)
                    if not ans:
                        raise Exception(body.get('error_msg', ''))

                # transforms user input
                transformations_methods = body.get('transformations', [])
                for trans in transformations_methods:
                    user_input = trans(user_input)

                # check for possible values
                body_values = body.get('values', None)
                if body_values:
                    body_values_keys = body['values'].keys()
                    if user_input in body_values_keys:
                        user_input = body['values'][user_input]
                    else:
                        raise Exception(body.get('error_msg', 'Oops something bad happened.'))
            break
        except Exception as e:
            click.secho(str(e), fg='red')

    return user_input



import inspect
import pandas as pd
import numpy as np
from sklearn.metrics import *
from sklearn.feature_selection import *
from sklearn.feature_selection import *
from sklearn.linear_model import *
from sklearn.ensemble import *
from sklearn.svm import *
from mlapp.utils.metrics.pandas import classification, regression, time_series
from sklearn.cluster import KMeans, MiniBatchKMeans, DBSCAN, MeanShift, SpectralClustering, AgglomerativeClustering
from sklearn.metrics import make_scorer
from sklearn.metrics._scorer import _BaseScorer
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.linear_model import LogisticRegression, Lasso, Ridge, BayesianRidge, ElasticNet, Lars, LassoLars, \
    LinearRegression
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier, ExtraTreesClassifier, \
    GradientBoostingRegressor, GradientBoostingClassifier, ExtraTreesRegressor, AdaBoostClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.svm import SVC
from mlapp.utils.exceptions.framework_exceptions import AutoMLException
from mlapp.utils.auto_ml.base import _AutoMLBase
from mlapp.utils.visualizations.pandas import visualization_methods

try:
    import xgboost as xgb
except ModuleNotFoundError as e:
    xgb = None
try:
    import lightgbm as lgb
except ModuleNotFoundError as e:
    lgb = None


class _AutoMLPandas(_AutoMLBase):
    def _validate_input(self, estimator_family, train, test, *args, **kwargs):
        if estimator_family not in ['linear', 'non_linear', 'binary', 'multi_class', 'clustering']:
            raise AutoMLException("ERROR: model family not supported.")

        for data in [train, kwargs['y_train'], test, kwargs['y_test']]:
            if not (isinstance(data, pd.DataFrame) or isinstance(data, pd.Series) or isinstance(data, np.ndarray)):
                raise AutoMLException("ERROR: data type should be Pandas Series/Dataframe or Numpy ndarray.")

        if train.shape[0] <= 1:
            raise AutoMLException("ERROR: Use more than 1 column for your train and test set!")

        if kwargs['search_type'] not in ['grid', 'random']:
            raise AutoMLException("ERROR: 'search_type' must be 'grid' or 'random'!")

        # TODO: add more validations tests

    def _eval_config_input(self, config_input):
        for key in config_input:
            if isinstance(config_input[key], dict):
                # go in one level
                self._eval_config_input(config_input[key])
            else:
                try:
                    # try to eval str
                    if isinstance(config_input[key], str):
                        config_input[key] = eval(config_input[key])
                except:
                    continue

        return config_input

    def _default_model_classes(self, estimator_family):
        return {
            'linear': {
                'Lasso': Lasso,
                'LinearRegression': LinearRegression,
                'Ridge': Ridge,
                'ElasticNet': ElasticNet,
                'Lars': Lars,
                'LassoLars': LassoLars,
                'BayesianRidge': BayesianRidge
            },
            'non_linear': {
                'RandomForest': RandomForestRegressor,
                'ExtraTree': ExtraTreesRegressor,
                'XGBoost': xgb.XGBRegressor if xgb else None,
                'LightGBM': lgb.LGBMRegressor if lgb else None,
                'GradientBoosting': GradientBoostingRegressor if not xgb and not lgb else None
            },
            'binary': {
                'Logistic': LogisticRegression,
                'RBF SVM': SVC,  # binary
                'MultiLogistic': LogisticRegression,
                'ExtraTreeClassifier': ExtraTreesClassifier,
                'RandomForest': RandomForestClassifier,
                'GaussianNB': GaussianNB,
                'GaussianProcess_Classifier': GaussianProcessClassifier,
                'KnearestNeighbors': KNeighborsClassifier,
                'XGBoostClassifier': xgb.XGBClassifier if xgb else None,
                'LightGBMClassifier': lgb.LGBMClassifier if lgb else None,
                'GradientBoostingClassifier': GradientBoostingClassifier if not xgb and not lgb else None,
                'AdaBoostClassifier': AdaBoostClassifier
            },
            'multi_class': {
                'Logistic': LogisticRegression,
                'RBF SVM': SVC,  # binary
                'MultiLogistic': LogisticRegression,
                'ExtraTreeClassifier': ExtraTreesClassifier,
                'RandomForest': RandomForestClassifier,
                'GaussianNB': GaussianNB,
                'GaussianProcess_Classifier': GaussianProcessClassifier,
                'KnearestNeighbors': KNeighborsClassifier,
                'XGBoostClassifier': xgb.XGBClassifier if xgb else None,
                'LightGBMClassifier': lgb.LGBMClassifier if lgb else None,
                'GradientBoostingClassifier': GradientBoostingClassifier if not xgb and not lgb else None,
                'AdaBoostClassifier': AdaBoostClassifier
            },
            'clustering': {
                'AgglomerativeClustering': AgglomerativeClustering,
                'KMeans': KMeans,
                'MiniBatchKmeans': MiniBatchKMeans,
                'DB-SCAN': DBSCAN,
                'MeanShift': MeanShift,
                'SpectralClustering': SpectralClustering
            }
        }.get(estimator_family, {})

    def _default_fixed_params(self, estimator_family):
        return {
            # TODO: add default fixed params
        }.get(estimator_family, {})

    def _default_hyper_params(self, estimator_family):
        return {
            # TODO: add default hyper params
        }.get(estimator_family, {})

    def _visualization_methods(self, method):
        return visualization_methods(method)

    def _get_model_intercept_and_coefficients(self, model, features):
        try:
            if len(model.coef_) == 1:
                coefficients = {features[index]: model.coef_[0][index] for index, feature in
                                enumerate(model.coef_[0])}
            else:
                coefficients = {features[index]: model.coef_[index] for index, feature in enumerate(model.coef_)}

            return model.intercept_, coefficients
        except:
            try:
                return 0, {feature: model.feature_importances_[index] for index, feature in enumerate(features)}
            except:
                try:
                    return 0, model.booster().get_fscore()
                except:
                    try:
                        return 0, model._Booster.get_fscore()
                    except:
                        print("Warning: not able to get model coefficients.")
                        return 0, []

    def _calc_cv_df(self, searches, key, cv_weights):
        # get splits
        splits = list(filter(lambda x: True if all(word in x for word in ['split', 'test_score']) else False,
                             searches[key].cv_results_.keys()))
        splits_df = pd.DataFrame(np.transpose([searches[key].cv_results_[x] for x in splits]), columns=splits)

        # create column for each hyper param
        hyper_params_attr = list(filter(
            lambda x: True if all(word in x for word in ['param_']) else False, searches[key].cv_results_.keys()))
        cv_results = pd.DataFrame()
        for column_name in hyper_params_attr:
            cv_results[column_name] = searches[key].cv_results_[column_name]

        # calculate hyper param score
        cv_results['max_score'] = splits_df.apply(lambda x: max(x), axis=1)
        cv_results['min_score'] = splits_df.apply(lambda x: min(x), axis=1)
        cv_results['mean_score'] = splits_df.apply(lambda x: np.mean(x), axis=1)
        cv_results['std_score'] = splits_df.apply(lambda x: np.std(x), axis=1)
        cv_results['estimator'] = key

        # add weighted mean score
        if not cv_weights:
            cv_weights = 1
        cv_results['weighted_mean_score'] = splits_df.apply(lambda x: np.mean(np.multiply(x, cv_weights)), axis=1)

        return cv_results

    def _get_cv_results(self, estimator_family, searches, features, scoring, greater_is_better, *args, **kwargs):
        """
        Function to get the best model in the GridSearchCV analysis.
        The analysis is based on the model that performs the best the score on the validation sets from the cross validation
        Indeed, during the cross validation, each set will be consecutively the validations set and obtains a score (from
        the scorer). Then, we average all those scores and the model that brings the higher (or lower) score is considered
        the best model.
        :param keys: all GS keys
        :param searches: actual results of the GS
        :param features: coefficients of the model
        :param sort_by: CV score to sort by (max_score, min_score, mean_score, std_score, weighted_mean_score)
        :param low_score_is_better: boolean - whether low score is better
        :param cv_weights: the weights of the CV splits
        :return: dataframe of features, name of the best model, instance of the best model, intercept, coefficients
        """
        sort_by = kwargs.get('sort_by', 'weighted_mean_score')

        # get cv results df
        df = pd.concat([self._calc_cv_df(searches, key, kwargs['cv_weights']) for key in searches.keys()], sort=False)\
            .sort_values(by=sort_by, ascending=False)

        # reorder columns
        columns = ['estimator', 'min_score', 'mean_score', 'weighted_mean_score', 'max_score', 'std_score']
        columns = columns + [c for c in df.columns if c not in columns]
        df = df[columns]

        # fetch first row results => best result
        best_model_key = df.reset_index()['estimator'][0]
        best_model = searches[df.reset_index()['estimator'][0]].best_estimator_
        best_cv_score = df.iloc[0]['mean_score']
        intercept, coefficients = self._get_model_intercept_and_coefficients(best_model, features)

        return df, best_model_key, best_model, best_cv_score, intercept, coefficients

    def _predict_train_test(self, model, train, test, *args, **kwargs):
        train_y_hat = model.predict(train).reshape(1, len(train))[0]
        test_y_hat = model.predict(test).reshape(1, len(test))[0]
        return train_y_hat, test_y_hat

    def _full_data_model_fit(self, best_model, train, test, *args, **kwargs):
        best_model.fit(pd.concat([train, test]), np.concatenate((kwargs['y_train'], kwargs['y_test'])))
        return best_model

    def _feature_selection_globals(self, method):
        if method in globals():
            return globals()[method]

    def _run_feature_selection(self, feature_selection, train, test, *args, **kwargs):
        feature_selection_clf = self._load_feature_selection_method(feature_selection.get('method', ''))
        evaluated_params = self._eval_config_input(feature_selection.get('params', {}))

        if not feature_selection_clf:
            return list(train.columns), train, test

        feature_selection_method = feature_selection_clf(**evaluated_params)
        feature_selection_method.fit(train, kwargs.get('y_train'))

        selected_features = list(train.columns[feature_selection_method.get_support()])
        return selected_features, train[selected_features], test[selected_features]

    def _get_scorer(self, estimator_family, metric=None, *args, **kwargs):
        if metric:
            if callable(metric):
                if isinstance(metric, _BaseScorer):
                    return metric
                else:
                    return make_scorer(metric, greater_is_better=kwargs.get('greater_is_better', True))
            else:
                return make_scorer(globals()[metric], greater_is_better=kwargs.get('greater_is_better', True))
        else:
            return {
                'linear': make_scorer(mean_squared_error, greater_is_better=False),
                'non_linear': make_scorer(mean_squared_error, greater_is_better=False),
                'multi_class': make_scorer(f1_score, average='weighted'),
                'binary': make_scorer(f1_score, average='binary')
            }.get(estimator_family)

    def _scores_function(self, scores_summary_func, estimator_family):
        # take scores summary by 'scores_summary' then by 'best_estimator'
        return {
            'classification': classification,
            'regression': regression,
            'time_series': time_series
        }.get(scores_summary_func, {
            'linear': regression,
            'non_linear': regression,
            'multi_class': classification,
            'binary': classification
        }.get(estimator_family))

    def _inner_search(self, estimator_family, train, test, model_key, model_class, fixed_params, hyper_params, scoring,
                      greater_is_better, cv, *args, **kwargs):
        model_object = model_class(**fixed_params.get(model_key, {}))
        tuned_params = hyper_params.get(model_key, {})
        auto_search_params = {
            'cv': cv,
            'verbose': 1,
            'refit': True,
            'scoring': self._get_scorer(estimator_family, scoring, greater_is_better=greater_is_better, *args, **kwargs),
            'n_jobs': kwargs.get('n_jobs', 1),
            'return_train_score': True
        }
        if kwargs['search_type'] == 'grid' or len(tuned_params.keys()) == 0:
            auto_search = GridSearchCV(model_object, tuned_params, **auto_search_params)
        else:
            auto_search = RandomizedSearchCV(model_object, tuned_params, **auto_search_params)

        auto_search.fit(train, kwargs['y_train'])

        return auto_search



from mlapp.managers import DataManager, pipeline
import pandas as pd
from mlapp.utils.features.pandas import extend_dataframe


class CrashCourseDataManager(DataManager):
    @pipeline
    def load_train_data(self, *args):
        data = pd.read_csv(self.data_settings["file_path"])
        return data

    @pipeline
    def clean_train_data(self, data):
        # Drop samples with missing label
        data.dropna(inplace=True, subset=[self.model_settings['target']])

        # Extract the list of features excluding the variable to predict
        features = list(data.columns)
        features.remove(self.model_settings['target'])

        # Calculate the mean value for each feature
        default_values_for_missing = data[features].mean(axis=0).to_dict()

        # Fill any missing values using the previous calculation
        data = data.fillna(default_values_for_missing)

        # Store the calculated missing values
        self.save_metadata('missing_values', default_values_for_missing)
        return data

    @pipeline
    def transform_train_data(self, data):
        # apply transformations
        data = self._transform_data(data)

        # save data frame
        self.save_dataframe('features', data)

        return data

    @pipeline
    def load_forecast_data(self,*args):
        data = pd.read_csv(self.data_settings["file_path"])
        return data

    @pipeline
    def clean_forecast_data(self, data):
        # get the missing values
        default_values_for_missing = self.get_metadata('missing_values', default_value={})

        # fill the missing values
        data = data.fillna(default_values_for_missing)
        return data

    @pipeline
    def transform_forecast_data(self, data):
        # apply transformations
        data = self._transform_data(data)
        return data

    def _transform_data(self, data):
        # extend data frame
        conf_extend_dataframe = self.data_settings.get('conf_extend_dataframe', {})
        data = extend_dataframe(
            data,
            y_name_col=self.model_settings.get('target'),
            lead_order=conf_extend_dataframe.get('lead', 0),
            lag_order=conf_extend_dataframe.get('lag', 0),
            power_order=conf_extend_dataframe.get('power', 0),
            log=conf_extend_dataframe.get('log', False),
            exp=conf_extend_dataframe.get('exp', False),
            sqrt=conf_extend_dataframe.get('sqrt', False),
            poly_degree=conf_extend_dataframe.get('interactions', 0),
            inverse=conf_extend_dataframe.get('inverse', False)
        )
        return data

    @pipeline
    def load_target_data(self, *args):
        raise NotImplementedError()



import time
import os
import json
import confluent_kafka
from mlapp.handlers.message_queues.message_queue_interface import MessageQueueInterface


class KafkaKerberosHandler(MessageQueueInterface):
    def __init__(self, settings):
        """
        Initializes the KafkaKerberosHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(KafkaKerberosHandler, self).__init__()

        self.config = {
            'bootstrap.servers': settings.get('hostnames', ''),
            'group.id': settings.get('group_id'),
            'session.timeout.ms': 6000,
            'enable.auto.commit': 'TRUE',
            # Your AD account principle name
            'sasl.kerberos.principal': '{0}@{1}'.format(settings.get('username', ''), settings.get('domain_name', '')),
            # Path to your keytab file
            'sasl.kerberos.keytab': settings.get('keytab'),
            'sasl.kerberos.service.name': 'kafka',
            'security.protocol': 'SASL_SSL',
            'sasl.mechanisms': 'GSSAPI',
            # Certificate used for creating JKS files
            'ssl.ca.location': settings.get('ca_location')
        }

    def send_message(self, queue_name, body):
        """
        Sends message to the queue
        :param queue_name: name of the topic/queue to send the message to
        :param body: message as string or bytes
        """
        producer = confluent_kafka.Producer(self.config)
        raw_bytes = json.dumps(body).encode('utf-8')
        producer.produce(queue_name, raw_bytes)
        producer.flush()

    def listen_to_queues(self, queue_names, callback):
        """
        Listen to queues/topics
        :param queue_names: list of queue/topic names to listen to
        :param callback: function to call upon receiving a message
        """
        consumer = confluent_kafka.Consumer(self.config)
        consumer.subscribe(queue_names)
        print('[*] Waiting for messages in ' + str(queue_names) + '. To exit press CTRL+C')

        while True:
            msg = consumer.poll()

            if msg is None:
                continue
            if msg.error():
                print("Consumer error: {}".format(msg.error()))
                continue

            callback(msg.value().decode('utf-8'))


from mlapp.handlers.wrappers.wrapper_interface import WrapperInterface


class SparkWrapper(WrapperInterface):
    def init(self):
        """
        Initializes the Wrapper for all handlers of `spark` type.
        """
        super(SparkWrapper, self).init('spark')

    def load_model(self, file_path, module):
        """
        Loads a spark model
        :param file_path: path to spark model file
        :param module: name of module to load
        """
        for handler_name in self._main_handlers:
            return self._handlers[handler_name].load_model(file_path, module)

    def exec_query(self, query, params=None):
        """
        Executes Query in the database.
        :param query: str - query to be executed.
        :param params: list - list of parameters to be used if necessary in query
        :return: result of query
        """
        for handler_name in self._main_handlers:
            self._handlers[handler_name].exec_query(query, params)

    def load_csv_file(self, file_path, sep=',', header=True, toPandas=False, **kwargs):
        """
        This function reads a csv file and return a spark DataFrame
        :param file_path: path to csv file
        :param sep: separator of csv file
        :param header: include header of file
        :param toPandas: to load as pandas DataFrame
        :param kwargs: other keyword arguments containing additional information
        :return: spark DataFrame (or pandas)
        """
        for handler_name in self._main_handlers:
            return self._handlers[handler_name].exec_query(
                file_path, sep=sep, header=header, toPandas=toPandas, **kwargs)


spark_instance = SparkWrapper()





from azureml.core.runconfig import RunConfiguration
from mlapp.integrations.aml.utils.env import create_env_from_requirements


def create_runconfig(aml_compute, env=None):
    # Create a new runconfig object
    aml_run_config = RunConfiguration()

    # Use the aml_compute you created above.
    aml_run_config.target = aml_compute

    if env:
        aml_run_config.environment = env
    else:
        aml_run_config.environment = create_env_from_requirements()

    return aml_run_config



import pandas as pd
import copy
from mlapp.utils.general import recursive_dict_update, search_key_recursive
from mlapp.utils.exceptions.base_exceptions import IoManagerException
from matplotlib.figure import Figure
try:
    import plotly.graph_objects as go
except:
    go = None


class IOManager(object):
    def __init__(self, ):
        """
        IO Manager constructor defines the structures of input or output that are common in an asset.
        e.g. dataframes, metadata (dictionaries) , images, objects
        """

        self.dataframes = {}
        self.tables = {}
        self.metadata = {}
        self.images = {}
        self.objects = {}
        # self.config = {}
        self.ids = {}

        self.structure = {
            'dataframes': self.dataframes,
            'tables': self.tables,
            'metadata': self.metadata,
            'images': self.images,
            'objects': self.objects,
            'ids': self.ids,

        }
        self.to_store = {
            'images': False,
            'objects': False,
            'dataframes': False,
            'tables': False,
            'metadata': False,
            # 'config': False
        }

    #### SET Functions

    def add_dataframe(self, name, data, to_table=None):
        """
        Adds a dataframe to the dictionary of dataframes
        :param name: dataframe name
        :param data: dataframe object
        :param to_table: table name in physical database, to be saved to
        :return:
        """
        if not self.dataframes.get(name, None) is None:
            self.dataframes[name] = pd.concat([self.dataframes[name], data])
        else:
            self.dataframes[name] = data
        if to_table:
            self.tables[name] = to_table

    def set_objects_value(self, objects_category, key, value):
        """
         Adds an object to the dictionary of objects, e.g. model pkl
        :param objects_category: category name defined by user, e.g. "models"
        :param key: object name
        :param value: object value to update
        :return:
        """
        if self.objects.get(objects_category, []):

            self.objects[objects_category][key] = value
        else:
            self.objects[objects_category] = {key: value}
        self.to_store['objects'] = True

    def add_objects(self, objects_category, objects_dict):
        """
        Replaces/set  dictionary of objects under certain category
        :param objects_category: category name
        :param objects_dict: dict of objects
        :return:
        """
        if not isinstance(objects_dict, dict):
            raise IoManagerException(
                'IOManager:add_objects() received non dictionary as input. please pass dictionary only')

        if self.objects.get(objects_category, []):
            self.objects[objects_category] = recursive_dict_update(self.objects[objects_category], objects_dict)
        else:
            self.objects[objects_category] = objects_dict

    def set_analysis_metadata_value(self, analysis_metadata_name, key, value):
        """

        :param analysis_metadata_name: name of the metadata dictionary
        :param key: internal key in the dictionary
        :param value: internal value in the dictionary
        :return:
        """

        if self.metadata.get(analysis_metadata_name, []):
            self.metadata[analysis_metadata_name][key] = value
        else:
            self.metadata[analysis_metadata_name] = {key: value}

    def add_analysis_metadata(self, analysis_metadata_name, metadata_dict):
        """
        :param analysis_metadata_name: name of the metadata dictionary
        :param metadata_dict:the metadata dictionary
        :return:
        """
        if not isinstance(metadata_dict, dict):
            raise IoManagerException(
                'IOManager:set_analysis_metadata() received non dictionary as input. please pass dictionary only')

        if self.metadata.get(analysis_metadata_name, []):
            self.metadata[analysis_metadata_name] = recursive_dict_update(self.metadata[analysis_metadata_name],
                                                                          metadata_dict)
        else:
            self.metadata[analysis_metadata_name] = metadata_dict

    def add_images(self, imgs):
        """
        :param imgs: dictionary of images with image names and objects
        :return:
        """
        if isinstance(imgs, dict):
            for key in imgs:
                if isinstance(imgs[key], list):
                    for index, figure in enumerate(imgs[key]):
                        self.add_image(key + '_' + str(index), figure)
                else:
                    self.add_image(key, imgs[key])
            if len(imgs):
                self.to_store['images'] = True
        else:
            raise IoManagerException("Error in `add_images`! Send a dictionary where: \n"
                            " - Keys are the names of the image/s \n"
                            " - Values are `matplotlib` Figures, or a list of `matplotlib` Figures.")

    def add_image(self, name, figure):
        """
        :param name: image name
        :param figure:  image object
        :return:
        """
        if isinstance(figure, Figure) or (go and isinstance(figure, go.Figure)):
            self.structure['images'][name] = figure
            self.to_store['images'] = True
        else:
            raise IoManagerException(f"{name} is not in the right format. " 
                                     "ML App supports only figures from the matplotlib or plotly.js library.")

    # def add_config(self, config):
    #     self.config = config
    #
    # def set_config_value(self, key, value):
    #     self.set_nested_value(self.config, '', key, value)

    def add_key(self, key, value):
        """
        Adding a new structure in the IO manager structure
        :param key: name of the structure
        :param value: value to be set
        :return:
        """
        self.structure[key] = value

    #### GET Functions

    def get_dataframes(self):
        """
        :return: returns dictionary of dataframes
        """
        return self.dataframes

    def get_tables(self):
        """
        :return: dictionary of dataframe name:table name
        """
        return self.tables

    def get_dataframe(self, name, default=None):
        """
        :param name: name of the dataframe to be retrieved
        :param default: default value in case it doesn't exist
        :return: dataframe or default value
        """
        return self.dataframes.get(name, default)

    def get_objects(self):
        """
        :return: returns dictionary of objects
        """
        return self.objects

    def get_objects_value(self, key, default=None):
        """

        :param key: object key to be returned
        :param default:
        :return:
        """
        return self.objects.get(key, default)

    def get_images_files(self):
        """
        :return: a dictionary of images
        """
        return self.images

    # def get_config(self):
    #     return copy.deepcopy(self.config)

    def get_metadata(self):
        """
        :return: dictionary of metadata
        """
        #return copy.deepcopy(self.metadata)
        return self.metadata

    def get_metadata_value(self, analysis_metadata_name, default=None):
        """

        :param analysis_metadata_name: name of metadataa dictionary
        :param default: default value in case it doesn't exist
        :return: dictionary
        """
        return self.metadata.get(analysis_metadata_name, default)

    def get_all_values(self):
        """

        :return: the IO Manager structure and its content
        """
        return copy.deepcopy(self.structure)

    def get_all_keys(self):
        """

        :return: the IO Manager structure
        """
        return list(self.structure.keys())

    def get_all_keys_per_cat(self, category):
        """
        :param category: the IO Manager specific category and all its values (e.g. objects)
        :return:
        """
        if category in self.structure.keys():
            return list(self.structure[category].keys())
        else:
            raise IoManagerException("Unknown nested category" + category + " in IO_Manager."
                                                                   "Please insert one of the categories supported:" + ", ".join(
                list(self.structure.keys())))

    def get_all_values_per_category(self, category):
        """
        :param category: the structure category name ( e.g. "objects")
        :return: the values in that specific category, if exists
        """
        if category in self.structure.keys():
            return self.structure[category]
        else:
            raise IoManagerException("Unknown nested category" + category + " in IO_Manager."
                                                                   "Please insert one of the categories supported:" + ", ".join(
                list(self.structure.keys())))


    def search_key_value(self,key):
        """
        searches a key in the whole structure
        :param key: key to search
        :return: all values found, in a list
        """
        results = []
        for res in search_key_recursive(self.structure,key):
            results.append(res)
        return results

    #########################
    @staticmethod
    def prepare_obj_for_json(to_json):
        return copy.deepcopy(to_json)

    # def prepare_models_objects_for_package(self):
    #     """
    #     Preparing result for pickle
    #     :return: result for pickle
    #     """
    #     response = copy.deepcopy(self.results['analysis_results'])
    #     for key in self.results['models_objects']:
    #         try:
    #             response[key] = copy.deepcopy(self.results['models_objects'][key])
    #         except AttributeError as e:
    #             print("Error copying object `{}`. Reason: ".format(key), str(e))
    #             print("Trying to copy using object's copy function..")
    #             if isinstance(self.results['models_objects'][key], list):
    #                 response[key] = [v.copy() for v in self.results['models_objects'][key]]
    #             if isinstance(self.results['models_objects'][key], dict):
    #                 response[key] = {k: v.copy() for (k, v) in self.results['models_objects'][key].items()}
    #             else:
    #                 response[key] = self.results['models_objects'][key].copy()
    #             print("Copy Success!")
    #     return response

    def set_nested_value(self, dictionary, category, key, value):
        if key in dictionary[category].keys():
            dictionary[category][key] = value
        else:
            for k, v in dictionary[category].items():
                self.set_nested_value(dictionary[category], k, key, value)

    # def get_value(self, dictionary,category, key):
    #     if category in dictionary.keys():
    #         if key in dictionary[category]:
    #             return copy.deepcopy(dictionary[category][key])
    #
    #         else:
    #             raise exception("Unknown key " + key + " in "+category+ "dictionary." +
    #                             "Please insert the key into the right dictionary before calling it ")
    #     else:
    #         raise exception("Unknown nested category: "+category)



import pandas as pd
from pyspark.ml.tuning import ParamGridBuilder
from pyspark.ml.regression import LinearRegression, GBTRegressor
from pyspark.ml.classification import LogisticRegression, LinearSVC, RandomForestClassifier, GBTClassifier
from pyspark.sql.dataframe import DataFrame
from pyspark.rdd import RDD
from pyspark.ml.evaluation import RegressionEvaluator, BinaryClassificationEvaluator
from multiprocessing.pool import ThreadPool
from pyspark.ml.tuning import CrossValidator, CrossValidatorModel
import pyspark.sql.functions as F
from pyspark.ml.classification import RandomForestClassifier
from pyspark.ml.feature import VectorSlicer, VectorAssembler, ChiSqSelector, MinMaxScaler
import numpy as np
from mlapp.utils.exceptions.framework_exceptions import AutoMLException
from mlapp.utils.auto_ml.base import _AutoMLBase
from mlapp.utils.metrics.spark import classification, regression
from mlapp.utils.metrics.spark import MapeEvaluator, F1ScoreEvaluator
from mlapp.utils.visualizations.spark import visualization_methods


class CrossValidationSpark(CrossValidator):
    """
    An implementation of spark cross validator that can accept various train/test indices methods (under development)
    """
    def __init__(self, estimator=None, estimatorParamMaps=None, evaluator=None, cv=3, seed=None, parallelism=1,
                 collectSubModels=False):
        # check if cv is a list or a tuple containing indices. This is a preparation for time series cross validation.
        if isinstance(cv, tuple):
            self.sequentialIndex = True
            numFolds = 1
        elif isinstance(cv, list):
            self.sequentialIndex = True
            numFolds = len(cv)
        else:
            self.sequentialIndex = False
            numFolds = cv

        super(CrossValidationSpark, self).__init__(estimator=estimator, estimatorParamMaps=estimatorParamMaps, evaluator=evaluator, numFolds=numFolds,
                 seed=seed, parallelism=parallelism, collectSubModels=collectSubModels)

    def _fit(self, dataset):
        est = self.getOrDefault(self.estimator)
        epm = self.getOrDefault(self.estimatorParamMaps)
        numModels = len(epm)
        eva = self.getOrDefault(self.evaluator)
        nFolds = self.getOrDefault(self.numFolds)
        seed = self.getOrDefault(self.seed)
        h = 1.0 / nFolds
        randCol = self.uid + "_rand"
        df = dataset.select("*", F.rand(seed).alias(randCol))
        metrics = np.zeros((numModels, nFolds))

        pool = ThreadPool(processes=min(self.getParallelism(), numModels))
        subModels = None
        collectSubModelsParam = self.getCollectSubModels()
        if collectSubModelsParam:
            subModels = [[None for j in range(numModels)] for i in range(nFolds)]

        for i in range(nFolds):
            if self.sequentialIndex:
                pass
                # todo pass a column name to base the split on. make sure the split conforms to sklearn norms.
                # idx = [1,2,3,4]
                # training.where(~col("id").isin(idx)).show()
            else:
                validateLB = i * h
                validateUB = (i + 1) * h
                condition = (df[randCol] >= validateLB) & (df[randCol] < validateUB)
                validation = df.filter(condition).cache()
                train = df.filter(~condition).cache()

            tasks = self._parallelFitTasks(est, train, eva, validation, epm, collectSubModelsParam)
            for j, metric, subModel in pool.imap_unordered(lambda f: f(), tasks):
                metrics[j,i] = metric
                if collectSubModelsParam:
                    subModels[i][j] = subModel

            validation.unpersist()
            train.unpersist()

        avgMetrics  = np.mean(metrics,axis=1)

        if eva.isLargerBetter():
            bestIndex = np.argmax(avgMetrics)
        else:
            bestIndex = np.argmin(avgMetrics)
        bestModel = est.fit(dataset, epm[bestIndex])
        return self._copyValues(CrossValidatorModel(bestModel, avgMetrics.tolist(), subModels)), metrics

    @staticmethod
    def _parallelFitTasks(est, train, eva, validation, epm, collectSubModel):
        modelIter = est.fitMultiple(train, epm)
        def singleTask():
            index, model = next(modelIter)
            metric = eva.evaluate(model.transform(validation, epm[index]))
            return index, metric, model if collectSubModel else None
        return [singleTask] * len(epm)


class _AutoMLSpark(_AutoMLBase):
    def _eval_config_input(self, config_input):
        for key in config_input:
            if isinstance(config_input[key], dict):
                # go in one level
                self._eval_config_input(config_input[key])
            else:
                try:
                    # try to eval str
                    if isinstance(config_input[key], str):
                        config_input[key] = eval(config_input[key])
                except:
                    continue

        return config_input

    def _validate_input(self, estimator_family, train, test, *args, **kwargs):
        if estimator_family not in ['linear', 'binary', 'multi_class']:
            raise AutoMLException("ERROR: model family not supported.")

        for data in [train, test]:
            if not (isinstance(data, DataFrame) or isinstance(data, RDD)):
                raise AutoMLException("ERROR: data type should be PySpark RDD or DataFrame.")

            if kwargs['variable_to_predict'] not in data.columns:
                raise AutoMLException(f"ERROR: variable_to_predict '{kwargs['variable_to_predict']}' "
                                      f"column missing in data input.")

        # TODO: add more validations tests

    def _visualization_methods(self, method):
        return visualization_methods(method)

    def _default_model_classes(self, estimator_family):
        return {
            'linear': {
                'LinearRegression': LinearRegression,
                'Lasso': LinearRegression,  # fixed_params sets elasticNetParams = 1, which makes this a Lasso model.
                'Ridge': LinearRegression,  # fixed_params sets elasticNetParams = 0, which makes this a Ridge model.
                'GBTRegressor': GBTRegressor
            },
            'binary': {
                'LogisticRegression': LogisticRegression,
                'SVC': LinearSVC,
                'RandomForestClassifier': RandomForestClassifier,
                'GBTClassifier': GBTClassifier
            },
            'multi_class': {
                'LogisticRegression': LogisticRegression,
                'RandomForestClassifier': RandomForestClassifier,
                'GBTClassifier': GBTClassifier
            }
        }[estimator_family]

    def _default_fixed_params(self, estimator_family):
        return {
            'linear': {
                'Lasso': {
                    'elasticNetParam': 1
                },
                'Ridge': {
                    'elasticNetParam': 0
                }
            }
            # TODO: add default fixed params
        }.get(estimator_family, {})

    def _default_hyper_params(self, estimator_family):
        return {
            # TODO: add default hyper params
        }.get(estimator_family, {})

    def _get_model_intercept_and_coefficients(self, model, features):
        try:
            return model.intercept, [list(c) for c in zip(features, list(model.coefficients))],
        except:
            try:
                return 0, [[features[model.featureImportances.indices[index]], value] for index, value in enumerate(model.featureImportances.values)]
            except:
                return 0, []

    def _calc_cv_df(self, model_key, metrics, param_map):
        new_param_map = []
        for d in param_map:
            new_dict = {}
            for k, v in d.items():
                new_dict[k.name] = v
            new_param_map.append(new_dict)
        res = pd.DataFrame(new_param_map)
        res.insert(0, 'std', np.std(metrics, axis=1))
        res.insert(0, 'mean_score', np.mean(metrics, axis=1))
        res.insert(0, 'estimator', model_key)
        return res

    def _get_cv_results(self, estimator_family, searches, features, scoring, greater_is_better, *args, **kwargs):
        keys = list(searches.keys())
        is_larger_better = self._get_scorer(estimator_family, scoring, *args, **kwargs).isLargerBetter()
        if is_larger_better:
            best_model_key = keys[int(np.argmax([max(searches[k][0].avgMetrics) for k in keys]))]
        else:
            best_model_key = keys[int(np.argmin([min(searches[k][0].avgMetrics) for k in keys]))]

        best_model = searches[best_model_key][0].bestModel
        best_cv_score = max(searches[best_model_key][0].avgMetrics)
        intercept, coefficients = self._get_model_intercept_and_coefficients(best_model, features)

        df = pd.concat([value[1] for value in searches.values()]).sort_values(
            by='mean_score', ascending=not greater_is_better)

        return df, best_model_key, best_model, best_cv_score, intercept, coefficients

    def _full_data_model_fit(self, *args, **kwargs):
        pass

    def _predict_train_test(self, model, train, test, *args, **kwargs):
        return model.transform(train), model.transform(test)

    def _feature_selection_globals(self, method):
        if method in globals():
            return globals()[method]

    def _run_feature_selection(self, feature_selection, train, test, *args, **kwargs):
        feature_selection_clf = self._load_feature_selection_method(feature_selection.get('method', ''))
        evaluated_params = self._eval_config_input(feature_selection.get('params', {}))

        features = [c for c in train.columns if c != kwargs['variable_to_predict']]
        assembler = VectorAssembler(inputCols=features, outputCol='features')
        v_train = assembler.transform(train)
        v_test = assembler.transform(test)

        v_train = v_train.select(['features', kwargs['variable_to_predict']])
        v_test = v_test.select(['features', kwargs['variable_to_predict']])

        if not feature_selection_clf:
            return features, v_train, v_test

        selector = feature_selection_clf(
            featuresCol='features', outputCol="selectedFeatures", **evaluated_params).fit(v_train)
        v_train = selector.transform(v_train).drop('features').withColumnRenamed("selectedFeatures", "features")
        v_test = selector.transform(v_test).drop('features').withColumnRenamed("selectedFeatures", "features")

        return [features[i] for i in selector.selectedFeatures], v_train, v_test

    def _get_scorer(self, estimator_family, metric=None, *args, **kwargs):
        variable_to_predict = kwargs['variable_to_predict']
        return {
            'areaUnderROC': BinaryClassificationEvaluator(metricName='areaUnderROC', labelCol=variable_to_predict),
            'areaUnderPR': BinaryClassificationEvaluator(metricName='areaUnderPR', labelCol=variable_to_predict),
            'f1_score': F1ScoreEvaluator(labelCol=variable_to_predict),
            'rmse': RegressionEvaluator(metricName='rmse', labelCol=variable_to_predict),
            'mse': RegressionEvaluator(metricName='mse', labelCol=variable_to_predict),
            'r2': RegressionEvaluator(metricName='r2', labelCol=variable_to_predict),
            'mae': RegressionEvaluator(metricName='mae', labelCol=variable_to_predict),
            'mape': MapeEvaluator(labelCol=variable_to_predict)
        }.get(metric, {
                'linear': RegressionEvaluator(metricName='mse', labelCol=variable_to_predict),
                'non_linear': RegressionEvaluator(metricName='mse', labelCol=variable_to_predict),
                'multi_class': F1ScoreEvaluator(labelCol=variable_to_predict),
                'binary': F1ScoreEvaluator(labelCol=variable_to_predict),
            }.get(estimator_family))

    def _scores_function(self, scores_summary_func, estimator_family):
        # take scores summary by 'scores_summary' then by 'best_estimator'
        return {
            'classification': classification,
            'regression': regression,
        }.get(scores_summary_func, {
            'linear': regression,
            'non_linear': regression,
            'multi_class': classification,
            'binary': classification
        }.get(estimator_family))

    def _inner_search(self, estimator_family, train, test, model_key, model_class, fixed_params, hyper_params, scoring,
                      greater_is_better, cv, *args, **kwargs):
        model_object = model_class(labelCol=kwargs['variable_to_predict'], **fixed_params.get(model_key, {}))
        tuned_parameters = hyper_params.get(model_key, {})

        # build hyper parameter grid:
        pgb = ParamGridBuilder()
        for tp_key in tuned_parameters:
            pgb = pgb.addGrid(getattr(model_object, tp_key), tuned_parameters[tp_key])
        param_map = pgb.build()

        # run cross validator:
        evaluator = self._get_scorer(estimator_family, scoring, *args, **kwargs)
        cv_f = CrossValidationSpark(
            estimator=model_object, estimatorParamMaps=param_map, evaluator=evaluator, cv=cv)

        hps, metrics = cv_f.fit(train)

        return hps, self._calc_cv_df(model_key, metrics, param_map)






import json
import os
os.environ['LC_ALL'] = 'en_US.UTF-8'
os.environ['LANG'] = 'en_US.UTF-8'
from mlapp.env_loader import EMPTY_STRING
from mlapp.handlers.message_queues.message_queue_interface import MessageQueueInterface
from mlapp.integrations.aml.utils.pipeline import run_pipeline_endpoint
from azureml.core import Run
from azureml.exceptions import RunEnvironmentException
from mlapp.utils.exceptions.framework_exceptions import SkipServiceException


class AMLQueue(MessageQueueInterface):
    def __init__(self, settings):
        """
        Initializes the AMLQueue with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(AMLQueue, self).__init__()

        # getting azureml workspace
        try:
            self.global_run = Run.get_context(allow_offline=False)
        except RunEnvironmentException as e:
            raise SkipServiceException('Skip AMLQueue handler')
        self.ws = self.global_run.experiment.workspace
        self.experiment_name = settings.get('experiment_name', self.global_run.experiment.name)
        if self.experiment_name == EMPTY_STRING:
            self.experiment_name = self.global_run.experiment.name

    def send_message(self, queue_name, body):
        """
        Sends message to the queue
        :param queue_name: name of the topic/queue to send the message to
        :param body: message as string or bytes
        """
        # submitting a new pipeline endpoint run
        run_pipeline_endpoint(
            workspace=self.ws,
            pipeline_endpoint_id=queue_name,
            experiment_name=self.experiment_name,
            config_str=body,
            pipeline_version=None
        )

    def listen_to_queues(self, queue_names, callback):
        """
        Listen to queues/topics - not relevant in AML Queue
        :param queue_names: list of queue/topic names to listen to
        :param callback: function to call upon receiving a message
        """
        raise NotImplemented()



import os
from mlapp.config import settings
from mlapp.integrations.aml.utils.experiment import run_local_compute_experiment
from mlapp.integrations.aml.utils.workspace import init_workspace

# init workspace
ws = init_workspace(
    settings['aml']['tenant_id'],
    settings['aml']['subscription_id'],
    settings['aml']['resource_group'],
    settings['aml']['workspace_name']
)

run_local_compute_experiment(ws, experiment_name='local_experiment', entry_script=os.path.join("deployment", 'aml_target_compute.py'))



import time
import datetime as dt
import uuid

from mlapp.managers.job_manager import JobManager
from mlapp.utils.exceptions.base_exceptions import ConfigKeyError, ConfigError, IoManagerException

TIME_FORMAT = '%Y-%m-%d %H:%M:%S'


class FlowManager(object):

    def __init__(self, job_id, config, **kwargs):
        """
        Class FlowManager -
        :param flow_job_id: unique job id
        :param config: json containing the model settings, data settings, job settings
        for each model in the flow, and another model config named "flow_config" and
        see documentation for example config
        :param handlers: can be initiated from the app
        :param kwargs: flow_id as a minimum, and any other input from predecessor steps as followings:
        {'input_from_predecessor':{'key':value,'key':value}}
        """

        if (config is None) or (config == '') or (config == {}):
            raise ConfigError("ModelFlow INIT ERROR: received an empty config")

        self.config = config

        self.has_flow_summary = True
        self.flow_config = self.config.get('flow_config', {})
        if self.flow_config == {}:
            self.has_flow_summary = False

        self.job_id = job_id  # this used to be the job_id

        self.flow_id = kwargs.get('flow_id', str(uuid.uuid4()))

        self.pipelines_configs = self.config.get('pipelines_configs', [])
        self.flow_name = config.get('job_settings', {}).get("flow_name", '')

        # If kwargs have an input from outside it should be stored in kwargs as followings:
        # {'input_from_predecessor':{'key':value,'key':value}}
        self.flow_data = {'jobs_outputs': kwargs.get('input_from_predecessor', [None])}
        self.jobs_results = {'models_results': []}

    def run(self):
        '''
        Runs the flow, passes the input from previous pipelines to the next ones.
        Each pipeline has its own job manager, its own run id , and is returning a dictionary output.
        The config can have instruction regarding input from predecessor keys and output keys to pass to the next pipeline.

        :return: status_code, run_ids - list of all pipelines run ids, pipeline_outputs - output from all pipelines
        '''
        if self.pipelines_configs is None:
            raise ConfigKeyError("'pipelines_configs' property is missing in your config file.'")
        print(">>>>>> Running flow" + self.flow_name + "...")
        run_ids = []
        pipelines_metadata = []
        for i in range(len(self.pipelines_configs)):
            job_config = self.pipelines_configs[i]
            job_flow_setting = job_config.get('flow_settings', {})

            start_time = time.strftime(TIME_FORMAT)
            asset_name = job_config['job_settings'].get('asset_name', i)
            asset_label = job_config['job_settings'].get('asset_label', '')
            print(">>>>>> Running model: {}...".format(asset_name))

            run_ids.append(str(uuid.uuid4()))
            pipelines_metadata.append({
                'run_id': run_ids[i],
                'asset_name': asset_name,
                'asset_label': asset_label
            })
            cur_job_manager = JobManager(self.job_id, job_config, **{'run_id': run_ids[i]})

            kwargs = {'input_from_predecessor': {}}

            if ('input_from_predecessor' in job_flow_setting) and (
                    len(job_flow_setting['input_from_predecessor']) > 0):
                for k in job_flow_setting['input_from_predecessor']:
                    try:
                        if i == len(self.flow_data['jobs_outputs'])-1:
                            kwargs['input_from_predecessor'][k] = self.flow_data['jobs_outputs'][i][k]
                        else:
                            kwargs['input_from_predecessor'][k] = self.flow_data['jobs_outputs'][i-1][k]
                    except KeyError as e:
                        print(e)
                        raise IoManagerException(
                            "Flow job #{}: Error: cannot find input key: {} in the "
                            "results dictionary from your model:{} due to: {}".format(i, k, asset_name, str(e)))

            cur_results = cur_job_manager.run_pipeline(**kwargs['input_from_predecessor'])
            self.jobs_results['models_results'].append(cur_results)
            return_dict = {}
            if 'return_value' in job_flow_setting:
                for output_key in job_flow_setting['return_value']:
                    try:
                        return_dict[output_key] = cur_results.search_key_value(output_key)
                    except:
                        raise IoManagerException(
                            'return value {} from asset: {} is not existing '.format(output_key, asset_name))
            self.flow_data['jobs_outputs'].append(return_dict)

            end_time = dt.datetime.strptime(
                time.strftime(TIME_FORMAT), TIME_FORMAT) - dt.datetime.strptime(start_time, TIME_FORMAT)
            print(">>>>>> It took me, {}.".format(end_time))

        # insert the results data into the data_settings config:
        # self.flow_config['data_settings']['flow_data'] = self.flow_data
        # self.flow_config['job_settings']['flow_models_ids'] = model_ids
        flow_run_id = str(uuid.uuid4())
        flow_job_manager = JobManager(self.job_id, self.flow_config, **{
            'run_id': flow_run_id,
            'has_flow_summary': self.has_flow_summary
        })
        if self.has_flow_summary:
            args = [self.flow_data['jobs_outputs'][1:], run_ids]
            flow_output = flow_job_manager.run_pipeline(*args)
            self.jobs_results['models_results'].append(flow_output)
            run_ids.append(flow_run_id)

        if len(run_ids) > 1:
            flow_job_manager.store_flow_metadata(self.flow_id, **{'pipelines_metadata': pipelines_metadata})

        print(">>>>>> Finished running flow.")

        # todo: make sure to return proper value or just return
        return 100, run_ids, self.flow_data['jobs_outputs'][1:]



import logging
from minio import Minio
from minio.error import S3Error
from mlapp.handlers.file_storages.file_storage_interface import FileStorageInterface


class MinioStorageHandler(FileStorageInterface):
    def __init__(self, settings):
        """
        Initializes the MinioStorageHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(MinioStorageHandler, self).__init__()
        configuration = settings

        if not configuration:
            logging.error('Configuration should be added to the file. Key should be "minio"')

        try:
            self.minioClient = Minio(configuration['endPoint'] + ":" + configuration['port'],
                                     access_key=configuration['accessKey'],
                                     secret_key=configuration['secretKey'],
                                     secure=configuration['secure'],
                                     region=configuration['region'])
            logging.info("File storage: Successful connection")
        except KeyError as e:
            logging.error("Missing parameter in file storage config %s" % str(e))

    def download_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Downloads file from file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        self.minioClient.fget_object(bucket_name, object_name, file_path)

    def upload_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Uploads file to file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        try:
            self.minioClient.make_bucket(bucket_name)
        except S3Error as e:
            if e.code == 'BucketAlreadyOwnedByYou':
                pass
            else:
                raise e
        try:
            # with open(from_file_path, 'rb') as from_file:
            self.minioClient.fput_object(bucket_name, object_name, file_path)
            logging.info('File storage: file %s successfully uploaded', object_name)
        except IOError as e:
            logging.error(e)

    def list_files(self, bucket_name, prefix="", *args, **kwargs):
        """
        Lists files in file storage
        :param bucket_name: name of the bucket/container
        :param prefix: prefix string to search by
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        files_names = []

        try:
            objects = self.minioClient.list_objects(bucket_name, prefix=prefix, recursive=True)
            for obj in objects:
                files_names.append(obj.object_name.encode('utf-8'))
        except Exception as e:
            logging.error(e)

        return files_names



import numpy as np

# train configurations
classification_config = {
    "pipelines_configs": [
        {
            "data_settings": {
                "model_name": "breast_cancer",
                "local_data_csvs": [{
                    "name": "breast_cancer",
                    "path": "data/breast_cancer.csv"
                }],
                # features handling configurations
                "variable_to_predict": "answer",
                "data_handling": {
                    # y variable feature engineering
                    "y_variable": {
                        "type": "binary",  # binary/multi/continuous
                        "categories_labels": ["NEGATIVE", "POSITIVE"],  # category labels
                        "continuous_to_category_bins": [-1, 1, 2],
                        # bins values in case there is a need to cut the y to categories
                        "label_to_predict": ["POSITIVE"]  # target label to predict
                    },
                    "features_for_filter": {},  # leave empty for not filter
                    "features_for_train": None,  # leave empty for selecting all features
                    "set_features_index": None,  # leave empty for indexing by row index
                    "features_to_remove": ["texture error", "area error", "smoothness error", "compactness error",
                                           "concave points error", "symmetry error", "worst smoothness",
                                           "worst compactness", "worst concavity", "worst concave points",
                                           "worst symmetry", "worst fractal dimension"],
                    "feature_remove_by_null_percentage": 0.3,
                    "dates_format": ["%d/%m/%Y", "%Y-%m-%d"],
                    "default_missing_value": 0,
                    "features_handling": {
                        "mean radius": {"fillna": "np.mean", "transformation": ["np.square", "np.sqrt"]},
                        "radius error": {"fillna": 0, "transformation": []}
                    },
                    "features_interactions": [],
                    "dates_transformation": {"extraction_date": "20180430", "columns": []},
                    "features_to_bin": [
                        {"name": "mean radius", "bins": [12.3, 15.3]},
                        {"name": "mean texture", "bins": [15, 23]},
                        {"name": "mean perimeter", "bins": [72, 109]},
                        {"name": "mean area", "bins": [361, 886]},
                        {"name": "mean smoothness", "bins": [0.074, 0.11]},
                        {"name": "mean compactness", "bins": [0.047, 0.137, 0.228]},
                        {"name": "mean concavity", "bins": [0.023, 0.12]},
                        {"name": "mean concave points", "bins": [0.025]},
                        {"name": "mean symmetry", "bins": [0.142, 0.2, 0.26]},
                        {"name": "mean fractal dimension", "bins": [0.0518, 0.0541, 0.0742, 0.0827]},
                        {"name": "radius error", "bins": [0.19, 0.56, 0.83]},
                        {"name": "worst area", "bins": [500, 1050]},
                        {"name": "worst perimeter", "bins": [85, 120]},
                        {"name": "worst texture", "bins": [16.6, 42]},
                        {"name": "worst radius", "bins": [12.5, 18]},
                        {"name": "fractal dimension error", "bins": [0.006, 0.0135]},
                        {"name": "concavity error", "bins": [0.011, 0.082, 0.15]},
                        {"name": "perimeter error", "bins": [1.3, 5]}
                    ],
                    "action_for_continuous_features": "auto_bin",  # auto_bin/keep_as_is/remove
                    # evaluator properties
                    "evaluator_settings": {
                        "filter_evaluator_threshold": 0.05,
                        "store_evaluator_features": True
                    }
                }
            },
            "model_settings": {},
            "job_settings": {
                "asset_name": "classification",
                "pipeline": "feature_engineering",
            }
        }
    ]
}



import os
import shutil

from mlapp.integrations.aml.utils.env import create_env_from_requirements
from mlapp.utils.general import create_directory, create_tempdir, delete_directory_with_all_contents
from azureml.core import Webservice, Run, Experiment
from azureml.core.model import InferenceConfig, Model
from azureml.core.webservice import AciWebservice
from azureml.exceptions import WebserviceException
from mlapp import MLApp
from mlapp.config import settings
from mlapp.managers.flow_manager import FlowManager
from mlapp.integrations.aml.utils.run_class import get_model_register_name
from mlapp.integrations.aml.utils.constants import AML_MLAPP_FOLDER


def deploy_model(
        ws, aci_service_name, experiment_name, asset_name, asset_label, run_id, cpu_cores, memory_gb, entry_script):
    env = create_env_from_requirements(endpoint=True)
    inference_config = InferenceConfig(source_directory=os.getcwd(),
                                       entry_script=entry_script,
                                       environment=env)

    deployment_config = AciWebservice.deploy_configuration(cpu_cores=cpu_cores, memory_gb=memory_gb)

    # model name
    model_name = get_model_register_name(run_id)
    try:
        model = Model(ws, name=model_name)
    except:
        # creating directory for download Model files for Model register
        tmp_path = create_tempdir(name='download_tmp')
        register_path = create_directory(AML_MLAPP_FOLDER, path=tmp_path)

        # getting RUN context
        experiment = Experiment(workspace=ws, name=experiment_name)
        tags = {"run_id": run_id, "asset_name": asset_name}
        if asset_label is not None:
            tags["asset_label"] = asset_label

        selected_run_id = None
        for run in Run.list(experiment, tags=tags, include_children=True, status='Completed'):
            run_metrics = run.get_metrics()
            exp_saved_run_id = run_metrics.get("run_id")
            if exp_saved_run_id == run_id:
                selected_run_id = run.id
                break
        if selected_run_id is None:
            raise Exception('ERROR: there is no matching Run object that associated with the run id %s in this experiment.' % str(run_id))
        current_run = Run(experiment=experiment, run_id=selected_run_id)

        # download files from run object
        current_run.download_files(output_directory=register_path)

        # register model
        model = Model.register(
            ws,
            model_path=register_path,
            model_name=model_name,
            tags=tags,
            description=asset_name)

        # deletes tmp dir and all content
        delete_directory_with_all_contents(tmp_path)

    # deploy model
    service = None
    try:
        service = Webservice(ws, name=aci_service_name)
        service.update(models=[model], inference_config=inference_config)
    except WebserviceException as e:
        if service:
            service.delete()
        service = Model.deploy(ws, aci_service_name, [model], inference_config, deployment_config)

    service.wait_for_deployment(True)


def get_best_model_in_experiment(ws, experiment_name, asset_name, asset_label, score_metric, greater_is_better):
    best_score_run_id = None
    best_score = None

    tags = {
        'asset_name': asset_name
    }
    if asset_label is not None:
        tags['asset_label'] = asset_label

    # for run in experiment.get_runs(tags=tags, include_children=True):
    experiment = ws.experiments[experiment_name]
    for run in Run.list(experiment, tags=tags, include_children=True, status='Completed'):
        run_metrics = run.get_metrics()
        run_details = run.get_details()
        # each logged metric becomes a key in this returned dict
        run_score = run_metrics.get(score_metric)
        if run_score is not None:
            run_id = run_metrics["run_id"]

            if best_score is None:
                best_score = run_score
                best_score_run_id = run_id
            else:
                if greater_is_better:
                    if run_score > best_score:
                        best_score = run_score
                        best_score_run_id = run_id
                else:
                    if run_score < best_score:
                        best_score = run_score
                        best_score_run_id = run_id

    if not best_score:
        raise Exception(f"Error: score metric '{score_metric}' was not found in any run!")

    if not best_score_run_id:
        raise Exception(f"Error: haven't found a run with score metric '{score_metric}' score metric.")

    print("Best model run_id: " + best_score_run_id)
    print("Best model score: " + str(best_score))

    return best_score_run_id


def insert_model_id(configuration, model_id):
    job_settings = configuration['pipelines_configs'][0]['job_settings']
    if 'model_id' not in job_settings:
        job_settings['model_id'] = model_id


def preprocess_deployment(model_path):
    output_dir = os.path.join(os.getcwd(), 'output')
    os.makedirs(output_dir, exist_ok=True)
    run_id = None

    mlapp_files_path = os.path.join(model_path, 'mlapp')

    for file_name in os.listdir(mlapp_files_path):
        # TODO: fetch this from logger txt file
        if '.pkl' in file_name:
            run_id = str(file_name.split('.')[0]).split('_')[0]

        shutil.move(os.path.join(mlapp_files_path, file_name), output_dir)

    return run_id


def get_predictions_path(run_ids):
    for run_id in run_ids:
        for file_name in os.listdir('output'):
            if 'predictions' in file_name and run_id in file_name:
                return os.path.join('output', file_name)




from abc import ABCMeta, abstractmethod


class FileStorageInterface:
    __metaclass__ = ABCMeta

    @abstractmethod
    def download_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Downloads file from file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def stream_file(self, bucket_name, object_name, *args, **kwargs):
        """
        Streams file from file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        :return: file stream
        """
        raise NotImplementedError()

    @abstractmethod
    def upload_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Uploads file to file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def list_files(self, bucket_name, prefix="", *args, **kwargs):
        """
        Lists files in file storage
        :param bucket_name: name of the bucket/container
        :param prefix: prefix string to search by
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        :return: file names list
        """
        raise NotImplementedError()





import json
from ast import literal_eval
from azureml.core import Run
from mlapp.integrations.aml.utils.constants import MAX_STRING_LENGTH


def load_config_from_string(config_str):
    try:
        config_str = config_str.decode("utf-8")
    except AttributeError as attrError:
        pass  # message body is string and not bytes

    print("Hello, The following task is consumed:   " + str(config_str))
    try:
        config_str = json.loads(literal_eval(config_str))
    except Exception as first_error:
        try:
            config_str = json.loads(config_str)
        except Exception as second_error:
            print("Error response: " + str('Message not in JSON format'))
            raise second_error

    return config_str


def get_model_register_name(run_id):
    return run_id[:MAX_STRING_LENGTH]


def tag_and_log_run(config):
    run = Run.get_context()
    pipeline_configs = []

    for job in config.get('pipelines_configs', [{}]):
        pipeline_configs.append(job.get('job_settings', {}))

    if 'flow_config' in config:
        pipeline_configs.append(config['flow_config'].get('job_settings', {}))

    for job_config in pipeline_configs:
        for key in ['pipeline', 'asset_name', 'asset_label']:
            value = job_config.get(key)
            if value is not None:
                value_str = value
                if isinstance(value, list):
                    value_str = ','.join(value)
                run.log(key, value=value_str)
                run.tag(key, value=value_str)


def tag_and_log_outputs(run_ids):
    run = Run.get_context()

    for run_id in run_ids:
        run.log("run_id", run_id)
        run.tag("run_id", run_id)




from mlapp.mlapp_cli.common.cli_utilities import to_lower, clean_spaces, is_int, to_int, list_one_check

add_services_options = {
    "mysql": {
        "MLAPP_SERVICE_TYPE": 'mysql',
        "MAIN_DB": {
            "display_name": "Is it your main database",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        "MYSQL_HOSTNAME": {
            "display_name": "Hostname",
            "short_description": "Press enter to set hostname as localhost",
            "default": '127.0.0.1',
            "required": True
        },
        "MYSQL_PORT": {
            "display_name": "Port",
            "short_description": 'Press enter to set port to 3306',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Port should contain digits only.",
            "default": 3306,
            "required": True
        },
        "MYSQL_DATABASE_NAME": {
            "display_name": "Database name",
            "short_description": 'Enter your database name',
            "required": True
        },
        "MYSQL_USER_ID": {
            "display_name": "Username",
            "short_description": 'Enter your database user id',
            "required": True
        },
        "MYSQL_PASSWORD": {
            "display_name": "Password",
            "short_description": 'Enter your database password'
        }
    },
    "postgres": {
        "MLAPP_SERVICE_TYPE": 'postgres',
        "MAIN_DB": {
            "display_name": "Is it your main database",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        "POSTGRES_HOSTNAME": {
            "display_name": "Hostname",
            "short_description": "Press enter to set hostname as localhost",
            "default": '127.0.0.1',
            "required": True
        },
        "POSTGRES_PORT": {
            "display_name": "Port",
            "short_description": 'Press enter to set port to 5432',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Port should contain digits only.",
            "default": 5432,
            "required": True
        },
        "POSTGRES_DATABASE_NAME": {
            "display_name": "Database name",
            "short_description": 'Enter your database name',
            "required": True
        },
        "POSTGRES_USER_ID": {
            "display_name": "Username",
            "short_description": 'Enter your database user id',
            "required": True
        },
        "POSTGRES_USE_IAM": {
            "display_name": "Is your connection uses AWS IAM authentication?",
            "short_description": 'Y/N, default no',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'false',
            "required": True,
            "save_user_input": True
        },
        'POSTGRES_IAM_ACCESS_KEY': {
            "display_name": "Access key",
            "short_description": "Enter your IAM access key token",
            "preceding_key": "POSTGRES_USE_IAM",
            "preceding_key_response": 'true',
            "required": False
        },
        'POSTGRES_IAM_SECRET_KEY': {
            "display_name": "Secret key",
            "short_description": "Enter your IAM secret key token",
            "preceding_key": "POSTGRES_USE_IAM",
            "preceding_key_response": 'true',
            "required": False
        },
        "POSTGRES_PASSWORD": {
            "display_name": "Password",
            "short_description": 'Enter your database password',
            "preceding_key": "POSTGRES_USE_IAM",
            "preceding_key_response": 'false',
            "required": False

        }
    },
    "mssql": {
        "MLAPP_SERVICE_TYPE": 'mssql',
        "MAIN_DB": {
            "display_name": "Is it your main database",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        "MSSQL_HOSTNAME": {
            "display_name": "Hostname",
            "short_description": "Press enter to set hostname as localhost",
            "default": '127.0.0.1',
            "required": True
        },
        "MSSQL_PORT": {
            "display_name": "Port",
            "short_description": 'Press enter to set port to 1433',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Port should contain digits only.",
            "default": 1433,
            "required": True
        },
        "MSSQL_DATABASE_NAME": {
            "display_name": "Database name",
            "short_description": 'Enter your database name',
            "required": True
        },
        "MSSQL_USER_ID": {
            "display_name": "Username",
            "short_description": 'Enter your database user id',
            "required": True
        },
        "MSSQL_PASSWORD": {
            "display_name": "Password",
            "short_description": 'Enter your database password'
        }
    },
    "snowflake": {
        "MLAPP_SERVICE_TYPE": 'snowflake',
        "MAIN_DB": {
            "display_name": "Is it your main database",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        "SNOWFLAKE_ACCOUNT": {
            "display_name": "Account",
            "short_description": "Enter your account, i.e:  xxx.east-us-2.azure",
            "required": True
        },
        "SNOWFLAKE_USER": {
            "display_name": "Username",
            "short_description": 'Enter your database user id',
            "required": True
        },
        "SNOWFLAKE_PASSWORD": {
            "display_name": "Password (Press `Enter` to keep empty in case you're using an active directory)",
            "short_description": 'Enter your database password'
        },
        "SNOWFLAKE_DATABASE": {
            "display_name": "Database",
            "short_description": 'Enter your database name'
        },
        "SNOWFLAKE_SCHEMA": {
            "display_name": "Schema",
            "short_description": 'Enter your schema name'
        },
        "SNOWFLAKE_WAREHOUSE": {
            "display_name": "Warehouse",
            "short_description": 'Enter your warehouse name'
        },
        "SNOWFLAKE_ROLE": {
            "display_name": "Role",
            "short_description": 'Enter your role'
        }
    },
    "rabbitmq": {
        "MLAPP_SERVICE_TYPE": 'rabbitmq',
        "MAIN_MQ": {
            "display_name": "Is it your main message queue",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        'RABBITMQ_HOSTNAME': {
            "display_name": "Hostname",
            "short_description": "Press enter to set hostname as localhost",
            "default": 'amqp://127.0.0.1',
            "required": True
        },
        'RABBITMQ_PORT': {
            "display_name": "Port",
            "short_description": 'Press enter to set port to 5673',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Port should contain digits only.",
            "default": 5673,
            "required": True
        },
        # 'RABBITMQ_LISTEN_QUEUES_NAMES': {
        #     "display_name": "Listen queues names",
        #     "short_description": 'Press enter listening queues names, separated by comma',
        #     "transformations": [clean_spaces, list_one_check],
        #     "required": True
        # },
        # 'RABBITMQ_SEND_QUEUE_NAMES': {
        #     "display_name": "Send queues names",
        #     "short_description": 'Press enter sending queues names, separated by comma',
        #     "transformations": [clean_spaces, list_one_check],
        #     "required": True
        # },
        'RABBITMQ_CONNECTION_TIMEOUT': {
            "display_name": "Connection timeout",
            "short_description": 'Enter your connection timeout number in seconds, default 15',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Connection timeout should contain digits only.",
            "default": 15,
            "required": True
        }
    },
    "minio": {
        "MLAPP_SERVICE_TYPE": 'minio',
        "MAIN_FS": {
            "display_name": "Is it your main file storage",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        'MINIO_ENDPOINT': {
            "display_name": "Endpoint",
            "short_description": "Press enter to set endpoint as localhost",
            "default": '127.0.0.1',
            "required": True
        },
        'MINIO_ACCESS_KEY': {
            "display_name": "Access key",
            "short_description": "Enter your access key token",
            "required": True
        },
        'MINIO_SECRET_KEY': {
            "display_name": "Secret key",
            "short_description": "Enter your secret key token",
            "required": True
        },
        'MINIO_PORT': {
            "display_name": "Port",
            "short_description": 'Press enter to set port to 9000',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Port should contain digits only.",
            "default": 9000,
            "required": True

        },
        'MINIO_SECURE': {
            "display_name": "Secure",
            "short_description": 'Secure every bucket, Y/N, default no',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'false',
            "required": False

        },
        'MINIO_REGION': {
            "display_name": "Region",
            "short_description": "Enter your cluster region (Optional)",
            "required": False
        }
    },
    "azure-blob": {
        "MLAPP_SERVICE_TYPE": 'azure_blob',
        "MAIN_FS": {
            "display_name": "Is it your main file storage",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        'AZURE_BLOB_ACCOUNT_NAME': {
            "display_name": "account name",
            "short_description": "Enter your account name token",
            "required": True
        },
        'AZURE_BLOB_ACCOUNT_KEY': {
            "display_name": "account key",
            "short_description": "Enter your account key",
            "required": True
        }
    },
    "databricks": {
        "MLAPP_SERVICE_TYPE": 'databricks',
        "MAIN_SPARK": {
            "display_name": "Is it your main spark cluster",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        'DATABRICKS_ADDRESS': {
            "display_name": "Address",
            "short_description": "Press enter to set address as localhost",
            "default": '127.0.0.1',
            "required": True
        },
        'DATABRICKS_TOKEN': {
            "display_name": "Token",
            "short_description": "Enter your Databricks cluster authorization token",
            "required": True
        },
        'DATABRICKS_CLUSTER_ID': {
            "display_name": "Cluster ID",
            "short_description": 'Enter your Databricks cluster id',
            "required": True
        },
        'DATABRICKS_PORT': {
            "display_name": "Port",
            "short_description": 'Press enter to set port to 15001',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Port should contain digits only.",
            "default": 15001,
            "required": True
        },
        'DATABRICKS_ORGANIZATION_ID': {
            "display_name": "Organization ID",
            "short_description": 'Enter your Databricks organization id',
            "required": True
        }
    },
    "azure-service-bus": {
        "MLAPP_SERVICE_TYPE": 'azure_service_bus',
        "MAIN_MQ": {
            "display_name": "Is it your main message queue",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        'AZURE_SERVICE_BUS_HOSTNAME': {
            "display_name": "Host Name",
            "short_description": "Enter the Host Name",
            "required": True
        },
        'AZURE_SERVICE_BUS_SHARED_ACCESS_KEY_NAME': {
            "display_name": "Shared Access Key Name (Policy Name)",
            "short_description": "Enter the Policy Name",
            "required": True
        },
        'AZURE_SERVICE_BUS_SHARED_ACCESS_KEY': {
            "display_name": "Shared Access Key (Primary Key)",
            "short_description": "Enter the Primary Key",
            "required": True
        }
    },
    "kafka": {
        "MLAPP_SERVICE_TYPE": 'kafka',
        "MAIN_MQ": {
            "display_name": "Is it your main message queue",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        'KAFKA_HOSTNAME': {
            "display_name": "Hostname",
            "short_description": "Press enter to set hostname as localhost",
            "default": '127.0.0.1',
            "required": True
        },
        'KAFKA_PORT': {
            "display_name": "Port",
            "short_description": 'Press enter to set port to 9092',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Port should contain digits only.",
            "default": 9092,
            "required": True
        },
        'KAFKA_CONNECTION_TIMEOUT': {
            "display_name": "Connection timeout",
            "short_description": 'Enter your connection timeout number in seconds, default 15',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Connection timeout should contain digits only.",
            "default": 15,
            "required": True
        }
    },
    "boto": {
        "MLAPP_SERVICE_TYPE": 'boto',
        "MAIN_FS": {
            "display_name": "Is it your main file storage",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        'BOTO_ACCESS_KEY': {
            "display_name": "Access key",
            "short_description": "Enter your access key token",
            "required": True
        },
        'BOTO_SECRET_KEY': {
            "display_name": "Secret key",
            "short_description": "Enter your secret key token",
            "required": True
        }
    },
    "spark-local": {
        "MLAPP_SERVICE_TYPE": 'spark_local',
        "MAIN_SPARK": {
            "display_name": "Is it your main local spark cluster",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        }
    },
    "spark": {
        "MLAPP_SERVICE_TYPE": 'spark',
        "MAIN_SPARK": {
            "display_name": "Is it your main spark cluster",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        }
    },
    "azureml-model-storage": {
        "MLAPP_SERVICE_TYPE": 'azureml_model_storage',
        "MAIN_FS": {
            "display_name": "Is it your main azureml file storage",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        }
    },
    "azureml-run-storage": {
        "MLAPP_SERVICE_TYPE": 'azureml_run_storage',
        "MAIN_FS": {
            "display_name": "Is it your main azureml file storage",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        "AML_RUN_STORAGE_HOSTNAME": {
            "display_name": "Hostname",
            "short_description": "Press enter to set hostname as localhost",
            "default": '127.0.0.1',
            "required": True
        },
        "AML_RUN_STORAGE_PORT": {
            "display_name": "Port",
            "short_description": 'Press enter to set port to 5432',
            "validations": [is_int],
            "transformations": [to_int],
            "error_msg": "Port should contain digits only.",
            "default": 5432,
            "required": True
        },
        "AML_RUN_STORAGE_DATABASE_NAME": {
            "display_name": "Database name",
            "short_description": 'Enter your database name',
            "required": True
        },
        "AML_RUN_STORAGE_USER_ID": {
            "display_name": "Username",
            "short_description": 'Enter your database user id',
            "required": True
        },
        "AML_RUN_STORAGE_PASSWORD": {
            "display_name": "Password",
            "short_description": 'Enter your database password'
        }
    },
    "azureml-queue": {
        "MLAPP_SERVICE_TYPE": 'azureml_queue',
        "MAIN_MQ": {
            "display_name": "Is it your main message queue",
            "short_description": 'Y/N, default yes',
            "transformations": [to_lower, clean_spaces],
            "values": {
                'y': 'true',
                'n': 'false',
                'yes': 'true',
                'no': 'false'
            },
            "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
            "default": 'true',
            "required": True
        },
        'AML_EXPERIMENT_NAME': {
            "display_name": "experiment name",
            "required": False
        },
    }
}



from pyspark.sql import SparkSession, SQLContext, DataFrame
from pyspark import SparkConf
from mlapp.handlers.spark.spark_interface import SparkInterface


class SparkHandler(SparkInterface):

    def __init__(self, settings):
        """
        Initializes the SparkHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        conf = SparkConf()
        settings = {key: value for (key, value) in settings.items() if not value == ""}
        conf.setAll(settings.items())
        if settings.get("enable_hive", False):
            self.spark = SparkSession.builder.config(conf=conf).enableHiveSupport().getOrCreate()
        else:
            self.spark = SparkSession.builder.config(conf=conf).getOrCreate()

        # Get all database configuration
        self.spark_settings = settings
        self.driver = self.spark_settings.get('driver')
        self.connector_type = "jdbc"
        self.db_type = self.spark_settings.get('db_type')
        self.hostname = self.spark_settings.get('hostname')
        self.port = self.spark_settings.get('port')
        self.username = self.spark_settings.get('username')
        self.password = self.spark_settings.get('password')
        self.database_name = self.spark_settings.get('database_name')
        self.database_options = self.spark_settings.get('database_options')
        self.url = None
        if self.driver is not None and self.connector_type is not None and self.db_type is not None and \
           self.hostname is not None and self.port is not None and self.username is not None and \
           self.password is not None and self.database_name is not None and self.database_options is not None:
            self.__generate_url__()

    # def __del__(self):
    #     # Handler destructor
    #     self.close_connection()
    #
    # def close_connection(self):
    #     """
    #     Close the connection to spark
    #     """
    #     try:
    #         if (self.spark is not None) and (self.spark.stop is not None):
    #             self.spark.stop()
    #     except Exception as e:
    #         raise e

    def load_csv_file(self, file_path, sep=',', header=True, toPandas=False, **kwargs):
        """
        This function reads a csv file and return a spark DataFrame
        :param file_path: path to csv file
        :param sep: separator of csv file
        :param header: include header of file
        :param toPandas: to load as pandas DataFrame
        :param kwargs: other keyword arguments containing additional information
        :return: spark DataFrame (or pandas)
        """
        try:
            df = self.spark.read.csv(path=file_path, header=header, sep=sep, **kwargs)
            df = df.toPandas() if toPandas else df

            return df
        except Exception as e:
            raise e

    def load_model(self, file_path, module):
        """
        Loads a spark model
        :param file_path: path to spark model file
        :param module: name of module to load
        """
        model_class_name = file_path.split('.')[1]
        try:
            exec(module)
            return eval(model_class_name).load(file_path)
        except Exception as e:
            print(str(e))
            raise Exception("Missing import of `{}` in `spark_db_handler.py`!".format(model_class_name + 'Model'))

    def exec_query(self, query, params=None, **kwargs):
        """
        Executes Query in the database.
        :param query: str - query to be executed.
        :param params: list - list of parameters to be used if necessary in query
        :return: result of query
        """
        if params is not None:
            query = query.format(*params)

        return self.spark.sql(query)

    def __generate_url__(self):
        """
        This method creates the database url with all options for connect to the database
        :return: void
        """

        semicolon = ":"
        double_slash = "//"
        single_slash = "/"
        self.url = self.connector_type + semicolon + self.db_type + semicolon + double_slash + \
            self.hostname + semicolon + \
            single_slash + self.database_name + self.database_options



from mlapp.integrations.aml.utils.compute import get_or_create_compute_target
from mlapp.integrations.aml.utils.pipeline import create_mlapp_pipeline_step, run_pipeline_steps
from mlapp.integrations.aml.utils.runconfig import create_runconfig
import os


def run_script(ws, datastore, compute_target, vm_size='STANDARD_D2_V2', min_nodes=0, max_nodes=4):
    # init components
    compute_target = get_or_create_compute_target(ws, compute_target, vm_size, min_nodes, max_nodes)
    run_config = create_runconfig(compute_target)

    # create pipelines steps
    steps = create_mlapp_pipeline_step(
        compute_target, run_config,
        source_directory=os.getcwd(),
        entry_script=os.path.join("deployment", "aml_target_compute_run.py"))

    # publish pipeline endpoint
    run_pipeline_steps(ws, steps, experiment_name='pipeline_from_local')



from azureml.core import Datastore


def get_datastore(workspace, name='workspaceblobstore'):
    return Datastore.get(workspace, datastore_name=name)


def register_blob_datastore(subscription_id, resource_group, workspace, datastore_name, container_name, account_name,
                       account_key, set_as_default=True):

    datastore = Datastore.register_azure_blob_container(
        workspace=workspace,
        datastore_name=datastore_name,
        grant_workspace_access=True,
        container_name=container_name,
        account_name=account_name,
        account_key=account_key,
        subscription_id=subscription_id,
        resource_group=resource_group
    )
    if set_as_default:
        datastore.set_as_default()

    return datastore



from sklearn.model_selection import train_test_split
from mlapp.managers import ModelManager, pipeline
from mlapp.utils.automl import AutoMLPandas


class BasicRegressionModelManager(ModelManager):
    def __init__(self, *args, **kwargs):
        ModelManager.__init__(self, *args, **kwargs)

    @pipeline
    def train_model(self, data):
        variable_to_predict = self.model_settings.get('variable_to_predict', 'target')
        train_size = self.model_settings.get('train_percent', 0.8)

        # preparation for train
        X_train, X_test, y_train, y_test = train_test_split(
            data.drop(variable_to_predict, axis=1), data[variable_to_predict], train_size=train_size)

        # auto ml
        result = AutoMLPandas('linear').run(X_train, y_train, X_test, y_test)

        # save results
        self.save_automl_result(result)

    @pipeline
    def forecast(self, data):
        result = self.get_automl_result()

        # get best model
        model = result.get_best_model()

        # get selected features from feature selection algorithm
        selected_features_names = result.get_selected_features()

        # filter data according to selected features
        filtered_data = data[selected_features_names]

        # predicting
        predictions = model.predict(filtered_data)

        # store predictions
        self.add_predictions(data.index, predictions, [], 'forecast')

    @pipeline
    def refit(self, data):
        raise NotImplementedError()



import numpy as np

# train configurations
advanced_regression_config = {
    "pipelines_configs": [
        {
            "data_settings": {
                # load data
                "data_sources": {
                    # local file
                    "local": {
                        "file_paths": ["data/diabetes.csv"]  # local files
                    },
                    # database
                    # "db": {
                    #     "query": "SELECT * FROM diabetes",  # query to fetch data-m ,
                    #     "args": []  # arguments for query
                    # },
                    # s3 file store
                    # "s3": {
                    #     "buckets": {
                    #         "input": {  # bucket name
                    #             "file_keys": []  # file keys in bucket
                    #         }
                    #     }
                    # }
                },
                # features handling configurations
                "data_handling": {
                    "features_for_train": [],  # leave empty for selecting all features
                    "set_features_index": [],  # leave empty for indexing by row index
                    # features to remove
                    "features_to_remove": [
                        "age", "sex", "tc", "ldl", "age_sex", "sex_age", "age_hdl", "hdl_age", "sex_tc", "tc_sex",
                        "sex_ldl", "ldl_sex", "bmi_hdl", "hdl_bmi", "map_hdl", "hdl_map", "tc_hdl", "hdl_tc",
                        "hdl_tch", "tch_hdl", "hdl_ltg", "ltg_hdl", "hdl_glu", "glu_hdl",

                    ],
                    "feature_remove_by_null_percentage": 0.3,
                    # features transformations and missing value strategies
                    "features_handling": {
                        "bmi": {"fillna": "np.mean", "transformation": ["np.square"]},
                        "hdl": {"fillna": "np.mean", "transformation": ["np.square"]},
                        "tch": {"fillna": "np.mean", "transformation": ["np.square"]},
                        "glu": {"fillna": "np.mean", "transformation": ["np.square"]}
                    },
                    "features_to_bin": [
                        {"name": "tch", "bins": [-0.05, 0.03]},
                        {"name": "hdl", "bins": [0.07]},
                    ],
                    "interactions": True,
                },
            },
            "model_settings": {
                "train_percent": 0.8,
                "variable_to_predict": "target",
                "auto_ml": {
                    "feature_selection": [
                        {
                            "method": "SelectKBest",
                            "params": {
                                'k': 7,
                                'score_func': 'f_regression'
                            }
                        },
                        {
                            "method": "RFE",
                            "params": {
                                "estimator": "SVR(kernel='linear')",
                                "n_features_to_select": 5,
                                "step": 1
                            }
                        },
                        {
                            "method": "AllFeatures"
                        }
                    ],
                    "estimators": {
                        "linear": {
                            "models": ["Lasso", "Ridge"]
                        },
                        "non_linear": {
                            "models": ["RandomForest",  "ExtraTree"],
                            "hyper_params": {
                                "RandomForest": {
                                    "max_depth": [1, 3],
                                    "n_estimators": [10, 50]
                                }
                            }
                        }
                    }
                }
            },
            "job_settings": {
                "asset_name": "advanced_regression",
                "pipeline": "train",
            }
        }
    ]
}


"""
Base Exception
MLApp Exception - inherit from Base Exception
"""


class MLAppBaseException(Exception):
    def __init__(self, message):
        self.message = message


class FrameworkException(MLAppBaseException):
    def __init__(self, message=None):
        if message is not None and isinstance(message, str):
            self.message = message

    def __str__(self):
        return "[ML APP ERROR] %s\n" % str(self.message)


class UserException(MLAppBaseException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[USER ERROR] %s\n" % str(self.message)


class FlowManagerException(UserException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[FLOW MANAGER ERROR] %s\n" % str(self.message)


class DataManagerException(UserException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[DATA MANAGER ERROR] %s\n" % str(self.message)


class ModelManagerException(UserException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[MODEL MANAGER ERROR] %s\n" % str(self.message)


class JobManagerException(UserException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[JOB MANAGER ERROR] %s\n" % str(self.message)


class PipelineManagerException(UserException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[PIPELINE MANAGER ERROR] %s\n" % str(self.message)


class EnvironmentException(UserException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[ENVIRONMENT ERROR] %s\n" % str(self.message)


class IoManagerException(FlowManagerException, DataManagerException, ModelManagerException, JobManagerException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[IO MANAGER ERROR] %s\n" % str(self.message)


class ConfigError(FlowManagerException, DataManagerException, ModelManagerException, JobManagerException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[CONFIG ERROR] %s\n" % str(self.message)


class ConfigKeyError(ConfigError):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[KEY ERROR] %s\n" % str(self.message)


class ConfigValueError(ConfigError):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "[VALUE ERROR] %s\n" % str(self.message)



cli_assets_help = {
    "create": \
'''Usage:

`mlapp assets create ASSET_NAME` - creates an asset named ASSET_NAME\n
`mlapp assets create ASSET_NAME -f` - forcing asset creation named ASSET_NAME (caution: can override existing asset)\n
''',
    "show": '''Show all your project assets.''',
    "rename": \
'''Usage:

`mlapp assets rename PREVIOUS_NAME NEW_NAME` - duplicates existing asset named PREVIOUS_NAME and renamed it NEW_NAME \n
`mlapp assets rename PREVIOUS_NAME NEW_NAME -d` - rename existing asset named PREVIOUS_NAME and renamed it NEW_NAME (deletes PREVIOUS_NAME asset) \n
'''

}

cli_boilerplates_help = {
    "install": \
'''Usage:

`mlapp boilerplates install BOILERPLATE_NAME` - install mlapp boilerplate named BOILERPLATE_NAME\n
`mlapp boilerplates install BOILERPLATE_NAME -f` - forcing boilerplate installation named BOILERPLATE_NAME (can override existing boilerplate)\n
`mlapp boilerplates install BOILERPLATE_NAME -r NEW_NAME` - install mlapp boilerplate named BOILERPLATE_NAME and rename it to NEW_NAME.
''',
    "show": '''Show all ML App's available boilerplates. ''',
    "rename": \
'''Usage:

`mlapp assets rename PREVIOUS_NAME NEW_NAME` - duplicates existing asset named PREVIOUS_NAME and renamed it NEW_NAME \n
`mlapp assets rename PREVIOUS_NAME NEW_NAME -d` - rename existing asset named PREVIOUS_NAME and renamed it NEW_NAME (deletes PREVIOUS_NAME asset) \n
'''

}

cli_services_help = {
    "add": \
'''Usage:

`mlapp services add SERVICE_NAME` - register a new service to ML App (hint: use `mlapp services show-types`
to see all ML App's available services).
''',
    "delete": \
'''Usage:

`mlapp services delete SERVICE_NAME` - deletes a registered service (hint: use `mlapp services show`
to see all your registered services).
''',
    "show": \
'''Usage:

`mlapp services show` - shows all your registered services.
''',
    "show_types": \
'''Usage:

`mlapp services show-types` - shows all ML App's available services.
''',

}

cli_environment_help = {
    "init": \
'''Usage:

`mlapp environment init` - creates an empty env file with default name '.env'.\n
`mlapp environment init NAME` - creates an empty env file named NAME.
''',
    "set": \
'''Usage:

`mlapp environment set NAME` - sets ML App to point on NAME environment file, (this command modifies config.py file).
''',
}

cli_cp_help = {
    "setup": \
'''Usgae:

`mlapp cp setup` - setup the control panel on your machine.
''',
    "start": \
'''Usgae:

`mlapp cp start` - starts docker-compose orchestration of the control panel, open chrome browser with the address 'http://localhost:8081'.
''',
    "stop": \
'''Usgae:

`mlapp cp stop` - stops the docker-compose orchestration for the control panel.
''',
}


import copy
import pandas as pd
from abc import ABC, abstractmethod
from mlapp.utils.auto_ml.pandas import _AutoMLPandas
try:
    from mlapp.utils.auto_ml.spark import _AutoMLSpark
except:
    _AutoMLSpark = None

from mlapp.utils.exceptions.framework_exceptions import AutoMLException


class AutoMLSettings:
    def __init__(self, metric=None, greater_is_better=True, search_type='grid', scores_summary=None):
        """
        AutoML Settings Object

        :param metric: scorer function for cross validation. By default auto-selects based on estimator type:
        'linear'/'non-linear': mean squared error
        'binary'/'multi_class': f1 score

        Accepts any scorer depending on framework. E.g `r2_score` from sklearn.metrics when using AutoMLPandas.

        :param greater_is_better: boolean, i.e. for MAPE, low is better.
        :param search_type: 'grid'/'random' - type of search to be run to find the best combination.
        :param scores_summary: 'regression'/'classification'/'time_series' or None auto-selects by estimator.
        """
        self.metric = metric
        self.scores_summary = scores_summary
        self.greater_is_better = greater_is_better
        self.search_type = search_type

    def is_same_settings(self, s: 'AutoMLSettings'):
        return (self.metric == s.metric or self.metric is s.metric) and \
               self.greater_is_better == s.greater_is_better and \
               self.scores_summary == s.scores_summary


class AutoMLResults:
    """
    AutoMLResults Object contains results of AutoML run/s.

    :param settings: AutoMLSettings Object
    """
    def __init__(self, settings=AutoMLSettings()):
        self.settings = settings
        self.cv_metrics = pd.DataFrame()
        self.best_cv_score = None
        self.best_model_metrics = {}
        self.best_model = None
        self.selected_features = []
        self.intercept = None
        self.coefficients = []
        self.figures = {}
        self.best_estimator = None
        self.best_model_key = None
        self.best_feature_selection = None
        self.best_train_predictions = None
        self.best_test_predictions = None
        self.scores_function = None

    def add_cv_run(self, scores_function, best_estimator, feature_selection, train_predicted, test_predicted,
                   selected_features, best_model, intercept, coefficients, cv_score, all_cv_scores, figures,
                   *args, **kwargs):

        # assign feature selection column
        if 'feature_selection' not in all_cv_scores.columns:
            all_cv_scores.insert(loc=1, column='feature_selection', value=feature_selection)

        self.cv_metrics = pd.concat([self.cv_metrics, all_cv_scores])   # append cv metrics

        # new best score
        if not self.best_cv_score or cv_score > self.best_cv_score:

            # overwrite best model results
            if kwargs.get('model_metrics'):
                self.best_model_metrics = kwargs['model_metrics']
            else:
                self.best_model_metrics = self._scores_summary(
                    scores_function, train_predicted, test_predicted, *args, **kwargs)

            self.best_model = best_model
            self.best_train_predictions = train_predicted
            self.best_test_predictions = test_predicted
            self.best_cv_score = cv_score
            self.intercept = intercept
            self.selected_features = selected_features
            self.coefficients = coefficients
            self.figures = figures
            self.best_model_key = best_model.__class__.__name__
            self.best_estimator = best_estimator
            self.best_feature_selection = feature_selection
            self.scores_function = scores_function

    def merge(self, results: 'AutoMLResults'):
        if not AutoMLResults.is_same_settings(self, results):
            raise AutoMLException("Error: Tried to merge AutoML runs results with different settings!")

        self.add_cv_run(None, results.best_estimator, results.best_feature_selection,
                        results.best_train_predictions, results.best_test_predictions,
                        results.selected_features, results.best_model, results.intercept, results.coefficients,
                        results.best_cv_score, results.cv_metrics, results.figures,
                        model_metrics=results.best_model_metrics)

    def get_best_model(self):
        """
        Gets best model from AutoML run.
        """
        return self.best_model

    def get_metrics(self):
        """
        Gets metrics result associated with the best model.
        """
        return copy.copy(self.best_model_metrics)

    def get_figures(self):
        """
        Gets figures produced by AutoML run associated with the best model.
        """
        return self.figures

    def print_report(self, full=True, ascending=False):
        """
        Prints score report of all runs done by the AutoML.

        :param full: prints scores from all model runs, or use just best representative from each model (True/False)
        :param ascending: show scores in ascending order (True/False)
        """
        with pd.option_context(
                'display.max_rows', None,
                'display.max_columns', None,
                'display.expand_frame_repr', False,
                'max_colwidth', -1):
            df = self.cv_metrics.sort_values(by='mean_score', ascending=ascending)
            if not full:
                df = df.drop_duplicates(['estimator'])
            print("------------------------------------ AutoML Report -------------------------------------------")
            print(df)
            print("----------------------------------------------------------------------------------------------")

    def get_metadata(self):
        """
        Gets all metadata associated with the best model:

        > scores: scores associated with estimator used
        > estimator_family: estimator family (linear/non-linear/classificaton/etc.)
        > model: best model class name
        > selected_features_names: selected features
        > intercept: intercept of model if exists
        > coefficients: coefficients / feature importance if exist
        > feature_selection: feature selection method
        > cv_score: cross validation score
        """
        return {
            "scores": self.get_metrics(),
            "estimator_family": self.best_estimator,
            "model_class_name": self.best_model_key,
            "selected_features_names": self.selected_features,
            "intercept": self.intercept,
            "coefficients": self.coefficients,
            "feature_selection": self.best_feature_selection,
            "cv_score": self.best_cv_score
        }

    def get_train_predictions(self):
        """
        Gets train predictions from the best model.
        """
        return self.best_train_predictions

    def get_test_predictions(self):
        """
        Gets test predictions from the best model.
        """
        return self.best_test_predictions

    def get_selected_features(self):
        """
        Gets selected features associated with the best model.
        """
        return self.selected_features

    def predict_proba_by_threshold(self, x_train, x_test, y_train, y_test, threshold):
        """
        Prints score report of all runs done by the AutoML.

        :param x_train: train data (pd.DataFrame)
        :param x_test: test data (pd.DataFrame)
        :param y_train: train data (pd.DataFrame/pd.Series)
        :param y_test: test data (pd.DataFrame/pd.Series)
        :param threshold: Threshold percentage/s to use for predict by probability ([float]/float)

        :return: dictionary containing relevant scores.
        """
        results = {}

        if hasattr(self.best_model, 'predict_proba'):
            train_predict_proba = self.best_model.predict_proba(x_train[self.selected_features])[:, 1]
            test_predict_proba = self.best_model.predict_proba(x_test[self.selected_features])[:, 1]

            if not isinstance(threshold, list):
                threshold = [threshold]

            for proba_threshold in threshold:
                train_predicted = train_predict_proba > proba_threshold
                test_predicted = test_predict_proba > proba_threshold
                results[proba_threshold] = self._scores_summary(
                    self.scores_function, self.best_train_predictions, self.best_test_predictions, y_train, y_test,
                    train_y_proba=train_predicted, test_y_proba=test_predicted)

            return results

        return "best model doesn't support 'predict_proba'"

    @staticmethod
    def is_same_settings(r1: 'AutoMLResults', r2: 'AutoMLResults'):
        return r1.settings.is_same_settings(r2.settings)

    @staticmethod
    def _scores_summary(scores_function, train_predicted, test_predicted, y_train=None, y_test=None, *args, **kwargs):
        if scores_function:
            return scores_function(train_predicted, test_predicted, y_train, y_test, *args, **kwargs)
        else:
            return {}


class _AutoMLWrapper(ABC):
    def __init__(self, estimator, settings=AutoMLSettings(), models=None, fixed_params=None, hyper_params=None,
                 model_classes=None, feature_selection=None, visualizations=None):
        if not isinstance(settings, AutoMLSettings):
            raise AutoMLException(f"Error: 'settings' must be of type: 'AutoMLSettings'.")

        self.estimator = estimator
        self.settings = settings
        self.models = models
        self.model_classes = model_classes
        self.fixed_params = fixed_params
        self.hyper_params = hyper_params
        self.feature_selection = feature_selection
        self.visualizations = visualizations
        self.auto_ml_class = None

    @abstractmethod
    def run(self, *args, **kwargs) -> 'AutoMLResults':
        pass

    def _run(self, x_train, x_test, *args, **kwargs):
        results = AutoMLResults(self.settings)
        return self.auto_ml_class().run(
            results, self.settings.scores_summary, self.feature_selection, self.estimator, x_train, x_test, self.models,
            self.fixed_params, self.hyper_params, self.model_classes, self.settings.metric,
            self.settings.greater_is_better, search_type=self.settings.search_type, visualizations=self.visualizations,
            *args, **kwargs)


class AutoMLPandas(_AutoMLWrapper):
    def __init__(self, estimator, settings=AutoMLSettings(), models=None, fixed_params=None, hyper_params=None,
                 model_classes=None, feature_selection=None, visualizations=None):
        """
        This algorithm enables data scientists to find the best models among different options with different
        parametrization. It will pick the best model among the models, including the hyper params combination
        that performed the configured scoring the best.

        :param settings: AutoMLSettings
        :param estimator: model family to use: 'linear', 'non_linear', 'binary', 'multi_class'.
        :param models: assets available in the functionality depending on estimator_family:

            linear:
                'Lasso' 'LinearRegression', 'Ridge', 'ElasticNet', 'Lars', 'LassoLars' and 'BayesianRidge'.

            non_linear:
                'RandomForest', 'ExtraTree', 'XGBoost', 'LightGBM' and 'GradientBoosting'.

            binary:
                'Logistic', 'RBF , 'MultiLogistic', 'ExtraTreeClassifier', 'RandomForest', 'GaussianNB',
                'GaussianProcess_Classifier', 'KnearestNeighbors', 'GradientBoosting', 'XGBoostClassifier',
                'LightGBMClassifier', 'GradientBoostingClassifier' and 'AdaBoostClassifier'.

            multi_class:
                'Logistic', 'RBF , 'MultiLogistic', 'ExtraTreeClassifier', 'RandomForest', 'GaussianNB',
                'GaussianProcess_Classifier', 'KnearestNeighbors', 'GradientBoosting', 'XGBoostClassifier',
                'LightGBMClassifier', 'GradientBoostingClassifier' and 'AdaBoostClassifier'.

            clustering:
                'AgglomerativeClustering', 'KMeans', 'MiniBatchKmeans', 'DB-SCAN', 'MeanShift' and 'SpectralClustering'.

        :param fixed_params: dictionary, initialize each model with these fixed params. Each key should correspond to a key
        from the assets and values should be a dictionary: {'param_name': 'param_value'}. By default, the assets will be
        run with default configuration. Please, refer to online documentation (sklearn ..) to find out what are the possible
        parameters and values.

        :param hyper_params: dictionary of hyper_params for the GridSearch explorations. Each key should correspond to a key
        from the assets and values should be a dictionary: {'param_name': [list_of_values_to_explore]}. By default, the
        assets will be run without any hyper_params. Please, refer to online documentation (sklearn ..) to find out
        what are the possible parameters and values.

        :param model_classes: dictionary of model classes to pass to be run by the AutoML.

        :param visualizations: dictionary where key is the name and value is the related function.
        Please refer to utils.visualizations to check which function are available.
        """
        if visualizations is None:
            visualizations = ["score_evolution_hyper_params"]

        super(AutoMLPandas, self).__init__(estimator, settings, models, fixed_params, hyper_params, model_classes,
                                           feature_selection, visualizations)

        self.auto_ml_class = _AutoMLPandas

    def run(self, x_train, y_train, x_test, y_test, cv=5, cv_weights=None, n_jobs=1, best_model_full_fit=False) \
            -> 'AutoMLResults':
        """
        Runs the AutoML

        :param x_train: pandas DataFrame of train set of features
        :param y_train: pandas Series or DataFrame train target.
        :param x_test: pandas DataFrame of test set of features
        :param y_test: pandas Series or DataFrame test target.
        :param cv: cross validation splits. By default, cv=None.
        :param cv_weights: array, weight for each split. By default, cv_weights=None.
        :param best_model_full_fit: boolean , whether to fit the best model to the whole data frame (train + test).
        By default, set to False.
        :param n_jobs: number of jobs to run in parallel. By default, n_jobs=1. If you are using XGBoost, n_jobs should
        stay equal to 1.

        :return: AutoMLResults Object
        """
        return self._run(x_train, x_test, y_train=y_train, y_test=y_test, cv=cv, cv_weights=cv_weights, n_jobs=n_jobs)


class AutoMLSpark(_AutoMLWrapper):
    def __init__(self, estimator, settings=AutoMLSettings(), models=None, fixed_params=None, hyper_params=None,
                 model_classes=None, feature_selection=None, visualizations=None):
        """
        This algorithm enables data scientists to find the best models among different options with different
        parametrization. It will pick the best model among the models, including the hyper params combination
        that performed the configured scoring the best.

        :param settings: AutoMLSettings
        :param estimator: model family to use: 'linear', 'binary', 'multi_class'.
        :param models: assets available in the functionality depending on estimator_family:

        linear:
            'LinearRegression' 'Lasso', 'Ridge', 'GBTRegressor'.

        binary:
            'LogisticRegression', 'SVC , 'RandomForestClassifier', 'GBTClassifier'.

        multi_class:
            'LogisticRegression' , 'RandomForestClassifier', 'GBTClassifier'

        :param fixed_params: dictionary, initialize each model with these fixed params. Each key should correspond to a
        key from the assets and values should be a dictionary: {'param_name': 'param_value'}. By default, the assets
        will be run with default configuration. Please, refer to online documentation (pyspark ..) to find out what are
        the possible parameters and values.

        :param hyper_params: dictionary of hyper_params for the GridSearch explorations. Each key should correspond to a
        key from the assets and values should be a dictionary: {'param_name': [list_of_values_to_explore]}. By default,
        the models will be run without any hyper_params. Please, refer to online documentation (pyspark ..) to find out
        what are the possible parameters and values.

        :param model_classes: dictionary of model classes to pass to be run by the AutoML.

        :param visualizations: dictionary where key is the name and value is the related function.
        Please refer to utils.visualizations to check which function are available.
        """
        super(AutoMLSpark, self).__init__(estimator, settings, models, fixed_params, hyper_params, model_classes,
                                          feature_selection, visualizations)

        if not _AutoMLSpark:
            raise AutoMLException("Error: install the `pyspark` library in order to use the AutoMLSpark.")

        self.auto_ml_class = _AutoMLSpark

    def run(self, train_data, test_data, variable_to_predict, cv=3) -> 'AutoMLResults':
        """
        Runs the AutoML

        :param train_data: pandas DataFrame of train set of features
        :param test_data: pandas DataFrame of test set of features
        :param variable_to_predict: column name of the variable to predict
        :param cv: cross validation splits. By default, cv=None.

        :return: AutoMLResults Object
        """
        return self._run(train_data, test_data, variable_to_predict=variable_to_predict, cv=cv)




from mlapp.handlers.databases.sql_alchemy_handler import SQLAlchemyHandler


class SnowflakeHandler(SQLAlchemyHandler):
    def __init__(self, settings):
        """
        Initializes the SnowflakeHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(SnowflakeHandler, self).__init__(settings)

        authenticator = '&authenticator=externalbrowser' if self.connections_parameters['password'] == '' else ''
        self.connection_string = 'snowflake://{user}:{password}@{account}/{database}/{schema}?' \
                                 'warehouse={warehouse}&role={role}{authenticator}'\
            .format(user=self.connections_parameters['user'],
                    password=self.connections_parameters['password'],
                    account=self.connections_parameters['account'],
                    database=self.connections_parameters['database'],
                    schema=self.connections_parameters['schema'],
                    warehouse=self.connections_parameters['warehouse'],
                    role=self.connections_parameters['role'],
                    authenticator=authenticator)



from mlapp.integrations.aml.utils.pipeline import run_pipeline_endpoint


def run_script(ws, pipeline_endpoint_name, experiment_name, config_str, pipeline_endpoint_version=None):
    # TODO: perhaps change config str to config path and load it as str
    # publish pipeline endpoint
    run_pipeline_endpoint(ws, pipeline_endpoint_name, experiment_name, config_str,
                          pipeline_version=pipeline_endpoint_version)


# import os
# os.environ['LC_ALL'] = 'en_US.UTF-8'
# os.environ['LANG'] = 'en_US.UTF-8'
# from mlapp.integrations.aml.utils.pipeline import run_pipeline_endpoint
# from mlapp.integrations.aml.utils.workspace import init_workspace
# from config import settings
# import json
#
# pipeline_id = 'pipeline_id'
# experiment_name = 'experiment_name'
# config = {}
# entry_script = 'deployment/aml_target_compute.py'
#
# workspace = init_workspace(
#     settings['aml']['tenant_id'],
#     settings['aml']['subscription_id'],
#     settings['aml']['resource_group'],
#     settings['aml']['workspace_name']
# )
#
# run_pipeline_endpoint(workspace, pipeline_id, experiment_name, json.dumps(config), entry_script)
#



import os, click, shutil, uuid, re
import numpy as np
from mlapp.utils.general import get_project_root
from mlapp.mlapp_cli.common.model_rename_dictionary import rename_dictionary
from mlapp.mlapp_cli.common.cli_utilities import create_file, copy_files, str_capitalize
from mlapp.mlapp_cli.cli_help import cli_boilerplates_help
try:
    from mlapp.mlapp_cli.cli_test_env import env_test
except ModuleNotFoundError:
    env_test = None


@click.group("boilerplates")
def commands():
    """
    MLApp Boilerplates Command

    Use it to install MLApp boilerplates.
    type --help on each boilerplates command to get more information.

    """
    pass


@commands.command("install", help=cli_boilerplates_help['install'])
@click.argument("name", required=True)
@click.option("-f", "--force", is_flag=True, default=False, help="Flag force will override existing asset_name file.")
@click.option('-r', '--new-name', default=None, help="Use it to rename an asset name on installation.")
def install(name, force, new_name):
    if env_test:
        available_assets = os.listdir(os.path.join(os.getcwd(), '..', 'assets'))
    else:
        available_assets = os.listdir(os.path.join(get_project_root(), 'assets'))

    if name is None:
        click.secho("Asset name is required, please try again.", fg='red')
    elif name in available_assets:
        src = os.path.join(get_project_root(), 'assets', name)
        dst = os.path.join(os.getcwd(), 'assets', name)
        try:
            if new_name is not None and isinstance(new_name, str):
                if os.path.exists(os.path.join(os.getcwd(), 'assets', new_name)) and not force:
                    click.secho('ERROR: asset "' + new_name + '" is already exists.', fg='red')
                    return
                token = str(uuid.uuid4())
                tmp_name = name + '_' + token
                dst = dst + '_' + token

            if force and new_name is None:
                shutil.rmtree(dst)
                shutil.copytree(src, dst)
            else:
                shutil.copytree(src, dst)

            if new_name is not None and isinstance(new_name, str):
                res = __rename__(name=name, new_name=new_name, dir_name=tmp_name, show_success_msg=False, force=force)
                shutil.rmtree(dst)
                if res:
                    return

            # add data files
            data_dest = os.path.join(os.getcwd(), 'data')
            include_list = {
                'crash_course': ['glass.csv', 'glass_forecast.csv'],
                'classification': ['breast_cancer.csv'],
                'spark_classification': ['breast_cancer.csv'],
                'basic_regression': ['diabetes.csv'],
                'spark_regression': ['diabetes.csv']
            }
            if name in include_list:
                os.makedirs(data_dest, exist_ok=True)
                copy_files(os.path.join(get_project_root(), 'data'), data_dest, include_list=include_list[name])

            click.secho(
                'Success: asset was installed successfully under the name: "' + (
                    new_name if new_name is not None and isinstance(
                        new_name, str) else name) + '"', fg='green')

        except Exception as e:
            click.secho(
                f"ERROR: asset \"{name}\" already exists in your project. \nHint: you can use the -r flag followed by NAME to rename asset or -f flag to force override (caution: -f flag may override existing project).", fg='red')
    else:
        click.secho(
            f"ERROR: couldn't find the asset \"{name}\".\nHint: run 'mlapp boilerplates show' to list all available assets.", fg='red')


@commands.command("show", help=cli_boilerplates_help['show'])
def show():
    if env_test:
        available_boilerplates = os.listdir(os.path.join(os.getcwd(), '..', 'assets'))
    else:
        available_boilerplates = os.listdir(os.path.join(get_project_root(), 'assets'))
    click.echo("Available boilerplates:")
    for boilerplate in available_boilerplates:
        if not boilerplate[0] == '_':
            click.echo('[*] ' + boilerplate)


# todo: duplicated for internal use. find a better solution
def __rename__(name, new_name, dir_name=None, show_success_msg=True, delete=False, force=False):
    if name is None or new_name is None:
        click.secho(
            "ERROR: asset old name and new name are required.\nHint: 'mlapp assets rename OLD_NAME NEW_NAME' is the right command format.", fg='red')
        return True

    if name == new_name:
        click.secho("ERROR: mlapp boilerplate old name and your asset new name are equal. please use a different name or dont use rename option.", fg='red')
        return True

    except_list = ["__pycache__"]
    if dir_name is None:
        old_name_path = os.path.join(os.getcwd(), 'assets', name)
    else:
        old_name_path = os.path.join(os.getcwd(), 'assets', dir_name)
    new_name_path = os.path.join(os.getcwd(), 'assets', new_name)

    if not os.path.exists(old_name_path):
        click.secho(
            f"ERROR: asset \"{name}\" is not exists.\nHint: use 'mlapp assets show' command to view all available assets.", fg='red')
    elif os.path.exists(new_name_path) and not force:
        click.escho(
            f"ERROR: asset \"{new_name}\" is already exists.", fg='red')
    else:

        if os.path.exists(new_name_path) and not force:
            # shutil.rmtree(new_name_path)
            # creates the new model directory
            os.mkdir(new_name_path)

        # base renaming
        if name in rename_dictionary.keys():
            model_dict = rename_dictionary.get(name, {})
            replace_files = model_dict.keys()
        else:
            model_dict = rename_dictionary.get('base', {})
            replace_files = model_dict.keys()

        # rename model name file by file.
        for rf in replace_files:
            try:
                file_additional_path = model_dict[rf].get('inner_path', '')

                except_list.append(name + rf)
                # files path
                if dir_name is None:
                    full_old_filename = os.path.join(os.getcwd(), "assets", name, file_additional_path,
                                                     name + rf)
                else:
                    full_old_filename = os.path.join(os.getcwd(), "assets", dir_name, file_additional_path,
                                                     name + rf)
                full_new_filename = os.path.join(os.getcwd(), "assets", new_name, file_additional_path, new_name + rf)

                # checks if needs to create sub directories.
                if not os.path.exists(os.path.join(os.getcwd(), "assets", new_name, file_additional_path)):
                    os.mkdir(os.path.join(os.getcwd(), "assets", new_name, file_additional_path))

                words_to_replace = model_dict[rf].get('words', [])

                # read file content
                file_content = ''''''
                with open(full_old_filename, 'r') as file:
                    lines = file.readlines()
                    for line in lines:
                        # line_striped = line.strip()

                        # replace new file path
                        for wr_d in words_to_replace:
                            word = wr_d.get('word', None)
                            word_type = wr_d.get('word_type', 'append-left')
                            word_format = wr_d.get('word_format', 'str.lower')
                            word_pattern = wr_d.get('word_pattern', None)
                            if word is not None:
                                # check word format and type
                                word_format_func = eval(word_format)
                                old_name_formatted = word_format_func(name)
                                new_name_formatted = word_format_func(new_name)

                                if word_type == 'append-left':
                                    if word_pattern is not None and isinstance(word_pattern, str):
                                        word_pattern = old_name_formatted + word_pattern
                                    elif word_pattern is not None and (
                                            isinstance(word_pattern, list) or isinstance(word_pattern, np.array)):
                                        wp_res = ''
                                        for wp in word_pattern:
                                            wp_res += old_name_formatted + wp + '|'
                                        word_pattern = wp_res[:-1]

                                    old_name_formatted = old_name_formatted + word
                                    new_name_formatted = new_name_formatted + word
                                else:
                                    # 'append-right'

                                    # check for patterns type
                                    if word_pattern is not None and isinstance(word_pattern, str):
                                        word_pattern += old_name_formatted
                                    elif word_pattern is not None and (
                                            isinstance(word_pattern, list) or isinstance(word_pattern, np.array)):
                                        wp_res = ''
                                        for wp in word_pattern:
                                            wp_res += wp + old_name_formatted + '|'
                                        word_pattern = wp_res[:-1]

                                    old_name_formatted = word + old_name_formatted
                                    new_name_formatted = word + new_name_formatted

                                # checking and replacing model names or using regex
                                if word_pattern is not None:
                                    line = re.sub(word_pattern, new_name_formatted, line)
                                elif old_name_formatted in line:
                                    line = line.replace(old_name_formatted, new_name_formatted)

                        file_content += line

                # creates the new file with the new name.
                create_file(file_name=full_new_filename, content=file_content)

            except Exception as e:
                shutil.rmtree(new_name_path)
                click.secho('ERROR: Oops something bad happened.', fg='red')
                return True

        # copy the rest of the files
        copy_files(old_name_path, new_name_path, except_list)

        if delete:
            shutil.rmtree(old_name_path)

        if show_success_msg:
            click.secho('Success: asset "' + name + '" was renamed to "' + new_name + '".', fg='green')



from mlapp.handlers.wrappers.wrapper_interface import WrapperInterface


class FileStorageWrapper(WrapperInterface):
    def init(self):
        """
        Initializes the Wrapper for all handlers of `file_storage` type.
        """
        super(FileStorageWrapper, self).init('file_storage')

    def download_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Downloads file from file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        for handler_name in self._main_handlers:
            self._handlers[handler_name].download_file(bucket_name, object_name, file_path, *args, **kwargs)

    def upload_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Uploads file to file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        for handler_name in self._main_handlers:
            self._handlers[handler_name].upload_file(bucket_name, object_name, file_path, *args, **kwargs)
    
    def stream_file(self, bucket_name, object_name, *args, **kwargs):
        """
        Streams file from file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        for handler_name in self._main_handlers:
            if hasattr(self._handlers[handler_name], 'stream_file'):
                return self._handlers[handler_name].stream_file(bucket_name, object_name, *args, **kwargs)

    def list_files(self, bucket_name, prefix="", *args, **kwargs):
        """
        Lists files in file storage
        :param bucket_name: name of the bucket/container
        :param prefix: prefix string to search by
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        files_names = []
        for handler_name in self._main_handlers:
            files_names.extend(self._handlers[handler_name].list_files(bucket_name, prefix=prefix, *args, **kwargs))
        return files_names

    def postprocessing(self):
        """
        This method is used to run postprocessing for different handlers.
        :return: None
        """
        for key in self._handlers:
            if hasattr(file_storage_instance._handlers[key], 'postprocess'):
                file_storage_instance._handlers[key].postprocess()


file_storage_instance = FileStorageWrapper()





from mlapp.handlers.databases.sql_alchemy_handler import SQLAlchemyHandler


class MySQLHandler(SQLAlchemyHandler):
    def __init__(self, settings):
        """
        Initializes the MySQLHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(MySQLHandler, self).__init__(settings)
        self.connection_string = 'mysql+pymysql://{0}:{1}@{2}:{3}/{4}'.format(
            self.connections_parameters['user_id'],
            self.connections_parameters['password'],
            self.connections_parameters['hostname'],
            str(self.connections_parameters['port']),
            self.connections_parameters['database_name'])




import sys


class Logger:
    def __init__(self, log, level, job_id):
        self.log = log
        self.level = level
        self.job_id = job_id

    def write(self, message):
        for line in message.rstrip().splitlines():
            self.log.log(self.level, line.rstrip())

        sys.__stdout__.write(message)

    def flush(self):
        pass




from mlapp.handlers.spark.spark_interface import SparkInterface
from livy import LivySession
from requests.auth import HTTPBasicAuth
from textwrap import dedent
import requests


class LivyHandler(SparkInterface):

    def __init__(self, settings):
        """
        Initializes the HiveHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        settings = {key: value for (key, value) in settings.items() if not value == ""}

        self.livy_settings = settings

        requests_session = requests.Session()
        requests_session.headers.update({'X-Requested-By': 'livy', 'Content-Type': 'application/json'})

        self.params = {
            "url": self.livy_settings.get('url'),
            "auth": HTTPBasicAuth(self.livy_settings.get('username'), self.livy_settings.get('password')),
            "driver_memory": self.livy_settings.get('driver_memory', "512m"),
            "driver_cores": self.livy_settings.get('driver_cores', 1),
            "executor_cores": self.livy_settings.get('executor_cores', 1),
            "executor_memory": self.livy_settings.get('executor_memory', "512m"),
            "num_executors": self.livy_settings.get('num_executors', 1),
            "queue": self.livy_settings.get('queue', "default"),
            "name": self.livy_settings.get('name', "mlapp"),
            "heartbeat_timeout": self.livy_settings.get('heartbeat_timeout', 60),
            "requests_session": requests_session
        }

    def exec_query(self, query, params=None, **kwargs):
        """
        Executes Query in the database.
        :param query: str - query to be executed.
        :param params: list - list of parameters to be used if necessary in query
        :return: result of query
        """
        df_variable_name = kwargs.get('df_variable_name', 'df')

        with LivySession.create(**self.params) as session:
            # Run some code on the remote cluster
            session.run(dedent("""
                spark.sql('use default')
                df = spark.sql('show databases')
            """.format(*params)))

            # Retrieve the result
            df = session.read(df_variable_name)

        return df



import urllib
from mlapp.handlers.databases.sql_alchemy_handler import SQLAlchemyHandler


class MssqlHandler(SQLAlchemyHandler):

    def __init__(self, settings):
        """
        Initializes the MssqlHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(MssqlHandler, self).__init__(settings)

        # preparing connection string
        connection_string = 'DRIVER={ODBC Driver 17 for SQL Server};' + \
                            'SERVER={0};DATABASE={1};UID={2};PWD={3};Port={4};'.format(
                                self.connections_parameters['hostname'],
                                self.connections_parameters['database_name'],
                                self.connections_parameters['user_id'],
                                self.connections_parameters['password'],
                                self.connections_parameters['port'])
        connection_string = urllib.parse.quote_plus(connection_string)
        self.query_placeholder = '?'
        self.connection_string = "mssql+pyodbc:///?odbc_connect=%s" % connection_string
